from gdsfactory.simulation.modes.find_neff_ng_dw_dh import find_neff_ng_dw_dh


def test_dw_dh(dataframe_regression):
    df = find_neff_ng_dw_dh(steps=3)
    dataframe_regression.check(df)


if __name__ == "__main__":
    test_dw_dh()



import copy as python_copy

import gdspy
from phidl.device_layout import CellArray, Device, DeviceReference

from gdsfactory.cell import avoid_duplicated_cells
from gdsfactory.component import Component, ComponentReference


def copy(
    D: Component, prefix: str = "", suffix: str = "_copy", cache: bool = True
) -> Device:
    """returns a deep copy of a Component.
    based on phidl.geometry with CellArray support
    """
    D_copy = Component(name=f"{prefix}{D.name}{suffix}")
    D_copy.info = python_copy.deepcopy(D.info)
    for ref in D.references:
        if isinstance(ref, DeviceReference):
            new_ref = ComponentReference(
                ref.parent,
                origin=ref.origin,
                rotation=ref.rotation,
                magnification=ref.magnification,
                x_reflection=ref.x_reflection,
            )
            new_ref.owner = D_copy
        elif isinstance(ref, gdspy.CellArray):
            new_ref = CellArray(
                device=ref.parent,
                columns=ref.columns,
                rows=ref.rows,
                spacing=ref.spacing,
                origin=ref.origin,
                rotation=ref.rotation,
                magnification=ref.magnification,
                x_reflection=ref.x_reflection,
            )
        D_copy.add(new_ref)
        for alias_name, alias_ref in D.aliases.items():
            if alias_ref == ref:
                D_copy.aliases[alias_name] = new_ref

    for port in D.ports.values():
        D_copy.add_port(port=port)
    for poly in D.polygons:
        D_copy.add_polygon(poly)
    for label in D.labels:
        D_copy.add_label(
            text=label.text,
            position=label.position,
            layer=(label.layer, label.texttype),
        )
    if cache:
        D_copy = avoid_duplicated_cells(D_copy)
    return D_copy



import pathlib
from dataclasses import asdict, field
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import pydantic
from phidl.device_layout import Device as Component

from gdsfactory.name import clean_value, get_name_short

module_path = pathlib.Path(__file__).parent.absolute()
Layer = Tuple[int, int]


def make_empty_dict() -> Dict[str, Callable]:
    return {}


@pydantic.dataclasses.dataclass(frozen=True)
class LayerMap:
    """Generic layermap based on Textbook:
    Lukas Chrostowski, Michael Hochberg, "Silicon Photonics Design",
    Cambridge University Press 2015, page 353

    You will need to create a new LayerMap with your specific foundry layers.
    """

    WG: Layer = (1, 0)
    WGCLAD: Layer = (111, 0)
    SLAB150: Layer = (2, 0)
    SLAB90: Layer = (3, 0)
    DEEPTRENCH: Layer = (4, 0)
    GE: Layer = (5, 0)
    WGN: Layer = (34, 0)
    WGN_CLAD: Layer = (36, 0)

    N: Layer = (20, 0)
    NP: Layer = (22, 0)
    NPP: Layer = (24, 0)
    P: Layer = (21, 0)
    PP: Layer = (23, 0)
    PPP: Layer = (25, 0)
    GEN: Layer = (26, 0)
    GEP: Layer = (27, 0)

    HEATER: Layer = (47, 0)
    M1: Layer = (41, 0)
    M2: Layer = (45, 0)
    M3: Layer = (49, 0)
    VIAC: Layer = (40, 0)
    VIA1: Layer = (44, 0)
    VIA2: Layer = (43, 0)
    PADOPEN: Layer = (46, 0)

    NO_TILE_SI: Layer = (71, 0)
    PADDING: Layer = (67, 0)
    DEVREC: Layer = (68, 0)
    FLOORPLAN: Layer = (99, 0)
    TEXT: Layer = (66, 0)
    PORT: Layer = (1, 10)
    PORTE: Layer = (69, 0)
    PORTH: Layer = (70, 0)
    LABEL: Layer = (201, 0)
    LABEL_SETTINGS: Layer = (202, 0)
    TE: Layer = (203, 0)
    TM: Layer = (204, 0)
    DRC_MARKER: Layer = (205, 0)
    LABEL_INSTANCE: Layer = (206, 0)
    ERROR_MARKER: Layer = (207, 0)
    ERROR_PATH: Layer = (208, 0)


LAYER = LayerMap()
PORT_MARKER_LAYER_TO_TYPE = {
    LAYER.PORT: "optical",
    LAYER.PORTE: "dc",
    LAYER.TE: "vertical_te",
    LAYER.TM: "vertical_tm",
}

PORT_LAYER_TO_TYPE = {
    LAYER.WG: "optical",
    LAYER.WGN: "optical",
    LAYER.SLAB150: "optical",
    LAYER.M1: "dc",
    LAYER.M2: "dc",
    LAYER.M3: "dc",
    LAYER.TE: "vertical_te",
    LAYER.TM: "vertical_tm",
}

PORT_TYPE_TO_MARKER_LAYER = {v: k for k, v in PORT_MARKER_LAYER_TO_TYPE.items()}


@pydantic.dataclasses.dataclass
class LayerLevel:
    """Layer For 3D LayerStack.

    Args:
        layer: (GDSII Layer number, GDSII datatype)
        thickness: layer thickness
        zmin: height position where material starts
        material: material name
        sidewall_angle: in degrees with respect to normal
    """

    layer: Tuple[int, int]
    thickness: Optional[float] = None
    zmin: Optional[float] = None
    material: Optional[str] = None
    sidewall_angle: float = 0


class LayerStack(dict):
    """For simulation and trimesh 3D rendering"""

    def get_layer_to_thickness(self) -> Dict[Tuple[int, int], float]:
        """Returns layer tuple to thickness (um)."""
        return {
            level.layer: level.thickness for level in self.values() if level.thickness
        }

    def get_layer_to_zmin(self) -> Dict[Tuple[int, int], float]:
        """Returns layer tuple to z min position (um)."""
        return {level.layer: level.zmin for level in self.values() if level.thickness}

    def get_layer_to_material(self) -> Dict[Tuple[int, int], str]:
        """Returns layer tuple to material name."""
        return {
            level.layer: level.material for level in self.values() if level.thickness
        }

    def get_layer_to_sidewall_angle(self) -> Dict[Tuple[int, int], str]:
        """Returns layer tuple to material name."""
        return {
            level.layer: level.sidewall_angle
            for level in self.values()
            if level.thickness
        }

    def to_dict(self) -> Dict[str, Dict[str, Any]]:
        return {level_name: asdict(level) for level_name, level in self.items()}


def get_layer_stack_generic(thickness_silicon_core: float = 220e-3) -> LayerStack:
    """Returns generic LayerStack.
    based on paper https://www.degruyter.com/document/doi/10.1515/nanoph-2013-0034/html
    """
    layer_stack = LayerStack(
        core=LayerLevel(
            layer=LAYER.WG,
            thickness=thickness_silicon_core,
            zmin=0.0,
            material="si",
        ),
        clad=LayerLevel(
            layer=LAYER.WGCLAD,
            zmin=0.0,
            material="sio2",
        ),
        slab150=LayerLevel(
            layer=LAYER.SLAB150,
            thickness=150e-3,
            zmin=0,
            material="si",
        ),
        slab90=LayerLevel(
            layer=LAYER.SLAB90,
            thickness=150e-3,
            zmin=0.0,
            material="si",
        ),
        nitride=LayerLevel(
            layer=LAYER.WGN,
            thickness=350e-3,
            zmin=220e-3 + 100e-3,
            material="sin",
        ),
        ge=LayerLevel(
            layer=LAYER.GE,
            thickness=500e-3,
            zmin=220e-3,
            material="ge",
        ),
        via_contact=LayerLevel(
            layer=LAYER.VIAC,
            thickness=1100e-3,
            zmin=90e-3,
            material="Aluminum",
        ),
        metal1=LayerLevel(
            layer=LAYER.M1,
            thickness=750e-3,
            zmin=220e-3 + 1100e-3,
            material="Aluminum",
        ),
        heater=LayerLevel(
            layer=LAYER.HEATER,
            thickness=750e-3,
            zmin=220e-3 + 1100e-3,
            material="TiN",
        ),
        viac=LayerLevel(
            layer=LAYER.VIA1,
            thickness=1500e-3,
            zmin=220e-3 + 1100e-3 + 750e-3,
            material="Aluminum",
        ),
        metal2=LayerLevel(
            layer=LAYER.M2,
            thickness=2000e-3,
            zmin=220e-3 + 1100e-3 + 750e-3 + 1.5,
            material="Aluminum",
        ),
    )
    return layer_stack


LAYER_STACK = get_layer_stack_generic()


@pydantic.dataclasses.dataclass
class Section:
    """
    Args:
        width: of the section
        offset: center to center
        layer:
        ports: optional name of the ports
        name: optional section name
        port_types:
    """

    width: float
    offset: float = 0
    layer: Layer = (1, 0)
    ports: Tuple[Optional[str], Optional[str]] = (None, None)
    name: Optional[str] = None
    port_types: Tuple[str, str] = ("optical", "optical")

    def __repr__(self):
        return "_".join(
            [
                f"{i}"
                for i in [
                    self.name,
                    int(self.width * 1e3),
                    self.layer[0],
                    self.layer[1],
                    self.ports[0],
                    self.ports[1],
                    self.port_types[0],
                    self.port_types[1],
                ]
                if i is not None
            ]
        )


@pydantic.dataclasses.dataclass
class SimulationSettings:
    """
    Args:
        background_material: for the background
        port_margin: on both sides of the port width (um)
        port_height: port height (um)
        port_extension: port extension (um)
        mesh_accuracy: 2 (1: coarse, 2: fine, 3: superfine)
        zmargin: for the FDTD region (um)
        ymargin: for the FDTD region (um)
        xmargin: for the FDTD region (um)
        wavelength_start: 1.2 (um)
        wavelength_stop: 1.6 (um)
        wavelength_points: 500
        simulation_time: (s) related to max path length 3e8/2.4*10e-12*1e6 = 1.25mm
        simulation_temperature: in kelvin (default = 300)
        frequency_dependent_profile: computes mode profiles for different wavelengths
        field_profile_samples: number of wavelengths to compute field profile

    """

    background_material: str = "sio2"
    port_margin: float = 1.5
    port_extension: float = 5.0
    mesh_accuracy: int = 2
    zmargin: float = 1.0
    ymargin: float = 3.0
    xmargin: float = 3.0
    wavelength_start: float = 1.2
    wavelength_stop: float = 1.6
    wavelength_points: int = 500
    simulation_time: float = 10e-12
    simulation_temperature: float = 300
    frequency_dependent_profile: bool = True
    field_profile_samples: int = 15

    mode_index: int = 0
    n_modes: int = 2
    thickness_pml: float = 1.0
    port_source_name: str = "o1"
    distance_source_to_monitors: float = 0.2
    mesh_step: float = 40e-3
    wavelength: float = 1.55


SIMULATION_SETTINGS = SimulationSettings()


def assert_callable(function):
    if not callable(function):
        raise ValueError(
            f"Error: function = {function} with type {type(function)} is not callable"
        )


@pydantic.dataclasses.dataclass
class Library:
    """Stores component factories for testing purposes.

    Args:
        name: of the factory
        factory: component name to function dict
        post_init: Optional function to run over component

    """

    name: str
    factory: Dict[str, Callable] = field(default_factory=make_empty_dict)
    post_init: Optional[Callable[[], None]] = None

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def get_component_names(self):
        return list(self.factory.keys())

    def register(
        self,
        function_or_function_list: Optional[
            Union[List[Callable[[], None]], Callable[[], None]]
        ] = None,
        **kwargs,
    ) -> None:
        """Registers component_factory function or functions into the factory."""
        function_or_function_list = function_or_function_list or []

        if not hasattr(function_or_function_list, "__iter__"):
            function = function_or_function_list
            assert_callable(function)
            self.factory[function.__name__] = function

        else:
            for function in function_or_function_list:
                assert_callable(function)
                self.factory[get_name_short(clean_value(function))] = function

        for function_name, function in kwargs.items():
            assert_callable(function)
            self.factory[function_name] = function

    def get_component(
        self,
        component: Union[str, Dict],
        **settings,
    ) -> Component:
        """Returns Component from library.
        Takes default settings from self.settings
        settings can be overwriten with kwargs
        runs any post_init functions over the component before it returns it.

        Priority (from lower to higher)
        - default in dataclass
        - component in case it's a dic
        - **settings

        Args:
            component:
            **settings
        """
        if isinstance(component, str):
            component = component
            component_settings = {}
        elif isinstance(component, dict):
            component_settings = component.copy()
            if "component" not in component_settings:
                raise ValueError(f"{component} is missing `component` key")

            component = component_settings.pop("component")
        else:
            raise ValueError(
                f"{component} needs to be a string or dict, got {type(component)}"
            )

        if component not in self.factory:
            raise ValueError(f"{component} not in {list(self.factory.keys())}")
        component = self.factory[component](**settings)
        if self.post_init and not hasattr(component, "_initialized"):
            self.post_init(component)
            component._initialized = True
        return component


@pydantic.dataclasses.dataclass
class Tech:
    name: str = "generic"
    layer: LayerMap = LAYER

    fiber_spacing: float = 50.0
    fiber_array_spacing: float = 127.0
    fiber_input_to_output_spacing: float = 200.0
    layer_label: Layer = LAYER.LABEL
    metal_spacing: float = 10.0

    sparameters_path: str = str(module_path / "gdslib" / "sparameters")
    simulation_settings: SimulationSettings = SIMULATION_SETTINGS


TECH = Tech()
LIBRARY = Library("generic_components")

if __name__ == "__main__":
    import gdsfactory as gf

    def mmi1x2_longer(length_mmi: float = 25.0, **kwargs):
        return gf.components.mmi1x2(length_mmi=length_mmi, **kwargs)

    def mzi_longer(**kwargs):
        return gf.components.mzi(splitter=mmi1x2_longer, **kwargs)

    ls = LAYER_STACK
    # print(ls.get_layer_to_material())
    # print(ls.get_layer_to_thickness())

    # s = Section(width=1, layer=(1, 0))
    # print(s)



try:
    from gdsfactory.autoplacer.chip_array import ChipArray
    from gdsfactory.autoplacer.library import Library

except ModuleNotFoundError:
    print(
        "klayout installation not found. You need to `pip install klayout` to use the klayout placer"
    )


def test_autplacer():

    lib = Library()
    mask = ChipArray("chip_array", 25e6, 25e6, 3, 4, lib)
    mask.pack_grid(lib.pop("align"))
    mask.pack_grid(lib.pop(".*"))
    mask.write("chip_array.gds")


if __name__ == "__main__":
    import gdsfactory as gf

    test_autplacer()
    gf.show("chip_array.gds")



from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.cross import cross
from gdsfactory.components.pad import pad_array
from gdsfactory.types import ComponentFactory, Optional


@cell
def add_fidutials(
    component: ComponentFactory = pad_array,
    gap: float = 50,
    left: Optional[ComponentFactory] = cross,
    right: Optional[ComponentFactory] = cross,
    top: Optional[ComponentFactory] = None,
    bottom: Optional[ComponentFactory] = None,
    **kwargs
) -> Component:
    """Return component with fidutials.

    Args:
        component: component to add to the new component.
        gap: from component to fidutial edge.
        left: optional left fidutial.
        right: optional right fidutial.
        top: optional top fidutial.
        bottom: optional bottom fidutial.

    """
    c = Component()
    component = component(**kwargs)
    r = c << component

    if left:
        x1 = c << left()
        x1.xmax = r.xmin - gap

    if right:
        x2 = c << right()
        x2.xmin = r.xmax + gap

    if top:
        y1 = c << top()
        y1.ymin = r.ymax + gap

    if bottom:
        y2 = c << bottom()
        y2.ymin = r.ymin - gap

    c.add_ports(r.ports)
    c.copy_child_info(component)
    return c


if __name__ == "__main__":
    c = add_fidutials(top=cross)
    c.show()



from typing import Optional, Tuple

import numpy as np
from numpy import ndarray

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.geometry.functions import DEG2RAD, extrude_path
from gdsfactory.tech import LAYER
from gdsfactory.types import Layer


def ellipse_arc(
    a: float,
    b: float,
    x0: float,
    theta_min: float,
    theta_max: float,
    angle_step: float = 0.5,
) -> ndarray:
    """Returns an elliptical arc.

    b = a *sqrt(1-e**2)

    An ellipse with a = b has zero eccentricity (is a circle)

    Args:
        a: ellipse semi-major axis
        b: semi-minor axis

    """
    theta = np.arange(theta_min, theta_max + angle_step, angle_step) * DEG2RAD
    xs = a * np.cos(theta) + x0
    ys = b * np.sin(theta)
    return np.column_stack([xs, ys])


def grating_tooth_points(
    ap: float,
    bp: float,
    xp: float,
    width: float,
    taper_angle: float,
    spiked: bool = True,
    angle_step: float = 1.0,
) -> ndarray:
    theta_min = -taper_angle / 2
    theta_max = taper_angle / 2

    backbone_points = ellipse_arc(ap, bp, xp, theta_min, theta_max, angle_step)
    if spiked:
        spike_length = width / 3
    else:
        spike_length = 0.0
    points = extrude_path(
        backbone_points,
        width,
        with_manhattan_facing_angles=False,
        spike_length=spike_length,
    )

    return points


def grating_taper_points(
    a: float,
    b: float,
    x0: float,
    taper_length: float,
    taper_angle: float,
    wg_width: float,
    angle_step: float = 1.0,
) -> ndarray:
    taper_arc = ellipse_arc(a, b, taper_length, -taper_angle / 2, taper_angle / 2)

    port_position = np.array((x0, 0))
    p0 = port_position + (0, wg_width / 2)
    p1 = port_position + (0, -wg_width / 2)
    points = np.vstack([p0, p1, taper_arc])
    return points


@gf.cell
def grating_coupler_elliptical(
    polarization: str = "te",
    taper_length: float = 16.6,
    taper_angle: float = 40.0,
    wavelength: float = 1.554,
    fiber_angle: float = 15.0,
    grating_line_width: float = 0.343,
    wg_width: float = 0.5,
    neff: float = 2.638,  # tooth effective index
    nclad: float = 1.443,
    layer: Tuple[int, int] = LAYER.WG,
    n_periods: int = 30,
    big_last_tooth: bool = False,
    layer_slab: Optional[Tuple[int, int]] = LAYER.SLAB150,
    slab_xmin: float = -1.0,
    slab_offset: float = 2.0,
    fiber_marker_width: float = 11.0,
    fiber_marker_layer: Optional[Layer] = gf.LAYER.TE,
    spiked: bool = True,
) -> Component:
    r"""Grating coupler with parametrization based on Lumerical FDTD simulation.

    Args:
        polarization: te or tm
        taper_length: taper length from input
        taper_angle: grating flare angle
        wavelength: grating transmission central wavelength (um)
        fiber_angle: fibre angle in degrees determines ellipticity
        grating_line_width
        wg_width: waveguide width
        neff: tooth effective index
        nclad: cladding effective index
        layer: LAYER.WG
        n_periods: number of periods
        big_last_tooth: adds a big_last_tooth
        layer_slab: layer that protects the slab under the grating
        slab_xmin: where 0 is at the start of the taper
        slab_offset:
        fiber_marker_width: width
        fiber_marker_layer: Optional circular marker
        spiked: grating teeth have sharp spikes to avoid non-manhattan drc errors


    .. code::

                      fiber

                   /  /  /  /
                  /  /  /  /

                _|-|_|-|_|-|___ layer
                   layer_slab |
            o1  ______________|

    """

    # Compute some ellipse parameters
    sthc = np.sin(fiber_angle * DEG2RAD)
    d = neff ** 2 - nclad ** 2 * sthc ** 2
    a1 = wavelength * neff / d
    b1 = wavelength / np.sqrt(d)
    x1 = wavelength * nclad * sthc / d

    a1 = round(a1, 3)
    b1 = round(b1, 3)
    x1 = round(x1, 3)

    period = a1 + x1

    c = gf.Component()
    c.info.polarization = polarization
    c.info.wavelength = wavelength

    # Make the taper
    p = taper_length / period
    a_taper = a1 * p
    b_taper = b1 * p
    x_taper = x1 * p

    x_output = a_taper + x_taper - taper_length
    pts = grating_taper_points(
        a=a_taper,
        b=b_taper,
        x0=x_output,
        taper_length=x_taper,
        taper_angle=taper_angle,
        wg_width=wg_width,
    )
    c.add_polygon(pts, layer)

    width = gf.snap.snap_to_grid(grating_line_width)
    gap = gf.snap.snap_to_grid(period - grating_line_width)

    xi = taper_length
    for p in range(n_periods):
        xi += gap + width / 2
        p = xi / period
        pts = grating_tooth_points(
            p * a1, p * b1, p * x1, width, taper_angle, spiked=spiked
        )
        c.add_polygon(pts, layer)
        xi += width / 2

    w = 1.0
    total_length = (
        period * n_periods
        + taper_length
        + grating_line_width / 2
        + period
        - grating_line_width
        + w / 2
    )

    if big_last_tooth:
        # Add last "large tooth" after the standard grating teeth

        a = total_length / (1 + x1 / a1)
        b = b1 / a1 * a
        x = x1 / a1 * a

        pts = grating_tooth_points(a, b, x, w, taper_angle, spiked=False)
        c.add_polygon(pts, layer)

    x = np.round(taper_length + x_output, 3)
    if fiber_marker_layer:
        circle = gf.components.circle(
            radius=fiber_marker_width / 2, layer=fiber_marker_layer
        )
        circle_ref = c.add_ref(circle)
        circle_ref.movex(x + fiber_marker_width / 2)

    name = f"vertical_{polarization.lower()}"

    c.add_port(
        name=name,
        midpoint=[x + fiber_marker_width / 2, 0],
        width=fiber_marker_width,
        orientation=0,
        layer=fiber_marker_layer,
        port_type=name,
    )

    c.add_port(
        name="o1", midpoint=[x_output, 0], width=wg_width, orientation=180, layer=layer
    )

    if layer_slab:
        slab_xmin += x_output + taper_length
        slab_length = total_length + slab_offset
        slab_width = (c.ysize + 2 * slab_offset) / 2
        c.add_polygon(
            [
                (slab_xmin, slab_width),
                (slab_length, slab_width),
                (slab_length, -slab_width),
                (slab_xmin, -slab_width),
            ],
            layer_slab,
        )

    return c


grating_coupler_elliptical_tm = gf.partial(
    grating_coupler_elliptical,
    grating_line_width=0.707,
    fiber_marker_layer=gf.LAYER.TM,
    polarization="tm",
    taper_length=30,
    slab_xmin=-2,
    neff=1.8,
    n_periods=16,
)


grating_coupler_elliptical_te = grating_coupler_elliptical


if __name__ == "__main__":
    # c = grating_coupler_elliptical_tm(taper_length=30)
    # c = grating_coupler_elliptical_te(layer_slab=None, with_fiber_marker=False)
    c = grating_coupler_elliptical(
        layer_slab=gf.LAYER.SLAB150, taper_length=50, slab_xmin=-5
    )
    # print(c.polarization)
    # print(c.wavelength)
    # print(c.ports)
    # c.pprint()
    c.show()



from pathlib import Path
from typing import Tuple

from omegaconf import DictConfig, OmegaConf

import gdsfactory as gf
from gdsfactory.mask.merge_markdown import merge_markdown
from gdsfactory.mask.merge_test_metadata import merge_test_metadata

# from gdsfactory.mask.merge_json import merge_json
from gdsfactory.mask.merge_yaml import merge_yaml
from gdsfactory.mask.write_labels import write_labels


def merge_metadata(
    gdspath: Path,
    labels_prefix: str = "opt",
    layer_label: Tuple[int, int] = gf.LAYER.TEXT,
) -> DictConfig:
    """Merges all mask metadata and returns test metadata
    This function works well only when you define the mask in YAML
    Exports a YAML file with only the cells information that have a valid test and measurement label

    For the cells that need to be measure we add test labels

    This is the automatic version of write_labels combined with merge_test_metadata

    .. code::

        CSV labels  -------
                          |--> merge_test_metadata dict
                          |
        YAML metatada  ----


    Args:
        gdspath: GDSpath
        labels_prefix
        layer_label: layer for the labels
    """
    mdpath = gdspath.with_suffix(".md")
    yaml_path = gdspath.with_suffix(".yml")
    test_metadata_path = gdspath.with_suffix(".tp.yml")

    build_directory = gdspath.parent.parent
    doe_directory = build_directory / "cache_doe"

    labels_path = write_labels(
        gdspath=gdspath, prefix=labels_prefix, layer_label=layer_label
    )

    # jsonpath = gdspath.with_suffix(".yml")
    # merge_json(doe_directory=doe_directory, jsonpath=jsonpath, **kwargs)

    mask_metadata = merge_yaml(doe_directory=doe_directory, yaml_path=yaml_path)
    merge_markdown(reports_directory=doe_directory, mdpath=mdpath)
    tm = merge_test_metadata(
        labels_prefix=labels_prefix,
        mask_metadata=mask_metadata,
        labels_path=labels_path,
    )

    test_metadata_path.write_text(OmegaConf.to_yaml(tm))
    return tm


if __name__ == "__main__":
    gdspath = (
        gf.CONFIG["samples_path"] / "mask_custom" / "build" / "mask" / "sample_mask.gds"
    )
    tm = merge_metadata(gdspath)



"""
Compute modes of a rectangular Si strip waveguide on top of oxide.
Note that you should only pay attention, here, to the guided modes,
which are the modes whose frequency falls under the light line --
that is, frequency < beta / 1.45, where 1.45 is the SiO2 index.

Since there's no special lengthscale here, you can just
use microns.  In general, if you use units of x, the frequencies
output are equivalent to x/lambda# so here, the freqeuncies will be
output as um/lambda, e.g. 1.5um would correspond to the frequency
1/1.5 = 0.6667.

"""
from typing import Dict

import meep as mp
import numpy as np
from meep import mpb

from gdsfactory.simulation.modes.disable_print import disable_print, enable_print
from gdsfactory.simulation.modes.get_mode_solver_rib import get_mode_solver_rib
from gdsfactory.simulation.modes.types import Mode, ModeSolverOrFactory

mpb.Verbosity(0)


def find_modes(
    mode_solver: ModeSolverOrFactory = get_mode_solver_rib,
    tol: float = 1e-6,
    wavelength: float = 1.55,
    mode_number: int = 1,
    parity=mp.NO_PARITY,
    **kwargs
) -> Dict[int, Mode]:
    """Computes effective index and group index for a mode.

    Args:
        mode_solver: function that returns mpb.ModeSolver
        tol: tolerance when finding modes
        wavelength: wavelength
        mode_number: mode order of the first mode
        paririty: mp.ODD_Y mp.EVEN_X for TE, mp.EVEN_Y for TM.

    Keyword Args:
        wg_width: wg_width (um)
        wg_thickness: wg height (um)
        slab_thickness: thickness for the waveguide slab
        ncore: core material refractive index
        nclad: clad material refractive index
        sy: simulation region width (um)
        sz: simulation region height (um)
        res: resolution (pixels/um)
        nmodes: number of modes

    Returns: Dict[mode_number, Mode]


    compute mode_number lowest frequencies as a function of k. Also display
    "parities", i.e. whether the mode is symmetric or anti_symmetric
    through the y=0 and z=0 planes.
    mode_solver.run(mpb.display_yparities, mpb.display_zparities)

    Above, we outputed the dispersion relation: frequency (omega) as a
    function of wavevector kx (beta).  Alternatively, you can compute
    beta for a given omega -- for example, you might want to find the
    modes and wavevectors at a fixed wavelength of 1.55 microns.  You
    can do that using the find_k function:
    """
    mode_solver = mode_solver(**kwargs) if callable(mode_solver) else mode_solver
    nmodes = mode_solver.nmodes
    omega = 1 / wavelength

    # Output the x component of the Poynting vector for mode_number bands at omega
    disable_print()
    k = mode_solver.find_k(
        parity,
        omega,
        mode_number,
        mode_number + nmodes,
        mp.Vector3(1),
        tol,
        omega * 2.02,
        omega * 0.01,
        omega * 10,
        # mpb.output_poynting_x,
        mpb.display_yparities,
        mpb.display_group_velocities,
    )
    enable_print()
    vg = mode_solver.compute_group_velocities()
    vg = vg[0]
    neff = np.array(k) * wavelength
    ng = 1 / np.array(vg)

    modes = {
        i: Mode(
            mode_number=i,
            neff=neff[index],
            wavelength=wavelength,
            ng=ng,
            E=mode_solver.get_efield(i),
            H=mode_solver.get_hfield(i),
            eps=mode_solver.get_epsilon().T,
            y=np.linspace(
                -1 * mode_solver.info["sy"] / 2,
                mode_solver.info["sy"] / 2,
                int(mode_solver.info["sy"] * mode_solver.info["res"]),
            ),
            z=np.linspace(
                -1 * mode_solver.info["sz"] / 2,
                mode_solver.info["sz"] / 2,
                int(mode_solver.info["sz"] * mode_solver.info["res"]),
            ),
        )
        for index, i in enumerate(range(mode_number, mode_number + nmodes))
    }

    return modes


if __name__ == "__main__":
    ms = get_mode_solver_rib(wg_width=0.5)
    m = find_modes(mode_solver=ms)

    # tol: float = 1e-6
    # wavelength: float = 1.55
    # mode_number: int = 1
    # parity = mp.NO_PARITY

    # mode_solver = get_mode_solver_rib()
    # nmodes = mode_solver.nmodes
    # omega = 1 / wavelength

    # # Output the x component of the Poynting vector for mode_number bands at omega
    # k = mode_solver.find_k(
    #     parity,
    #     omega,
    #     mode_number,
    #     mode_number + nmodes,
    #     mp.Vector3(1),
    #     tol,
    #     omega * 2.02,
    #     omega * 0.01,
    #     omega * 10,
    #     mpb.output_poynting_x,
    #     mpb.display_yparities,
    #     mpb.display_group_velocities,
    # )
    # enable_print()
    # vg = mode_solver.compute_group_velocities()
    # vg0 = vg[0]
    # neff = np.array(k) * wavelength
    # ng = 1 / np.array(vg0)

    # modes = {
    #     i: Mode(
    #         mode_number=i,
    #         neff=neff[index],
    #         solver=mode_solver,
    #         wavelength=wavelength,
    #     )
    #     for index, i in enumerate(range(mode_number, mode_number + nmodes))
    # }



import gdsfactory as gf


def test_partial_function_with_kwargs():
    mmi400 = gf.partial(gf.components.mmi1x2, width=0.4)
    mmi400_args = gf.partial(gf.components.mmi1x2, 0.4)
    mmi600 = gf.partial(gf.components.mmi1x2, width=0.6)
    mzi400 = gf.partial(gf.components.mzi, splitter=mmi400)
    mzi600 = gf.partial(gf.components.mzi, splitter=mmi600)

    c400 = mzi400()
    c600 = mzi600()

    assert c600.name != c400.name, f"{c600.name} must be different from {c400.name}"

    cmmi400 = mmi400()
    cmmi400_args = mmi400_args()
    assert (
        cmmi400_args.name == cmmi400.name
    ), f"{cmmi400_args.name} must be equal to {cmmi400.name}"


def test_partial_function_without_kwargs():
    r1 = gf.partial(gf.c.rectangle, size=(4, 2))
    r2 = gf.partial(gf.c.rectangle, size=(4, 2))
    r3 = gf.partial(gf.c.rectangle, (4, 2))

    c1 = r1()
    c2 = r2()
    c3 = r3()

    assert c1.name == c2.name == c3.name, f"{c1.name} == {c2.name} == {c3.name}"


if __name__ == "__main__":
    # test_partial_function_with_kwargs()
    test_partial_function_without_kwargs()
    # mmi400 = gf.partial(gf.components.mmi1x2, width=0.4)
    # mmi600 = gf.partial(gf.components.mmi1x2, width=0.6)
    # mzi400 = gf.partial(gf.components.mzi, splitter=mmi400)
    # mzi600 = gf.partial(gf.components.mzi, splitter=mmi600)
    # c400 = mzi400()
    # c600 = mzi600()
    # print(c400.name, c600.name)



"""Returns Component from YAML syntax.

name: myComponent
vars:
    length: 3

info:
    description: just a demo
    polarization: TE
    ...

instances:
    mzi:
        component: mzi_phase_shifter
        settings:
            delta_length: ${vars.length}
            length_x: 50

    pads:
        component: pad_array
        settings:
            n: 2
            port_names:
                - e4

placements:
    mzi:
        x: 0
    pads:
        y: 200
        x: mzi,cc
ports:
    o1: mzi,o1
    o2: mzi,o2


routes:
    electrical:
        links:
            mzi,etop_e1: pads,e4_0
            mzi,etop_e2: pads,e4_1

        settings:
            layer: [31, 0]
            width: 10
            radius: 10

"""

import functools
import hashlib
import io
import json
import pathlib
import warnings
from typing import IO, Any, Callable, Dict, List, Optional, Union

import numpy as np
import omegaconf
from omegaconf import OmegaConf

from gdsfactory.add_pins import add_instance_label
from gdsfactory.cell import CACHE
from gdsfactory.component import Component, ComponentReference
from gdsfactory.components import factory
from gdsfactory.cross_section import cross_section_factory
from gdsfactory.routing.factories import routing_strategy as routing_strategy_factories
from gdsfactory.types import ComponentFactoryDict, CrossSectionFactory, Route

valid_placement_keys = ["x", "y", "dx", "dy", "rotation", "mirror", "port"]


valid_top_level_keys = [
    "name",
    "instances",
    "placements",
    "connections",
    "ports",
    "routes",
    "vars",
    "info",
]

valid_anchor_point_keywords = [
    "ce",
    "cw",
    "nc",
    "ne",
    "nw",
    "sc",
    "se",
    "sw",
    "center",
    "cc",
]
# refer to an (x,y) Point

valid_anchor_value_keywords = [
    "south",
    "west",
    "east",
    "north",
]
# refer to a singular (x or y) value

valid_anchor_keywords = valid_anchor_point_keywords + valid_anchor_value_keywords
# full set of valid anchor keywords (either referring to points or values)

valid_route_keys = [
    "links",
    "settings",
    "routing_strategy",
]
# Recognized keys within a YAML route definition


def _get_anchor_point_from_name(
    ref: ComponentReference, anchor_name: str
) -> Optional[np.ndarray]:
    if anchor_name in valid_anchor_point_keywords:
        pt = getattr(ref.size_info, anchor_name)
        return pt
    elif anchor_name in ref.ports:
        return ref.ports[anchor_name].position
    else:
        return None


def _get_anchor_value_from_name(
    ref: ComponentReference, anchor_name: str, return_value: str
) -> Optional[float]:
    if anchor_name in valid_anchor_value_keywords:
        v = getattr(ref.size_info, anchor_name)
        return v
    else:
        anchor_point = _get_anchor_point_from_name(ref, anchor_name)
        if anchor_point is None:
            return None
        if return_value == "x":
            return anchor_point[0]
        elif return_value == "y":
            return anchor_point[1]
        else:
            raise ValueError("Expected x or y as return_value.")


def place(
    placements_conf: Dict[str, Dict[str, Union[int, float, str]]],
    connections_by_transformed_inst: Dict[str, Dict[str, str]],
    instances: Dict[str, ComponentReference],
    encountered_insts: List[str],
    instance_name: Optional[str] = None,
    all_remaining_insts: Optional[List[str]] = None,
) -> None:
    """Place instance_name with placements_conf config.

    Args:
        placements_conf: Dict of instance_name to placement (x, y, rotation ...)
        connections_by_transformed_inst: Dict of connection attributes.
            keyed by the name of the instance which should be transformed
        instances: Dict of references
        encountered_insts: list of encountered_instances
        instance_name: instance_name to place
        all_remaining_insts: list of all the remaining instances to place
            instances pop from this instrance as they are placed
    """
    if not all_remaining_insts:
        return
    if instance_name is None:
        instance_name = all_remaining_insts.pop(0)
    else:
        all_remaining_insts.remove(instance_name)

    if instance_name in encountered_insts:
        encountered_insts.append(instance_name)
        loop_str = " -> ".join(encountered_insts)
        raise ValueError(
            f"circular reference in placement for {instance_name}! Loop: {loop_str}"
        )
    encountered_insts.append(instance_name)
    ref = instances[instance_name]

    if instance_name in placements_conf:
        placement_settings = placements_conf[instance_name] or {}
        if not isinstance(placement_settings, omegaconf.DictConfig):
            raise ValueError(
                f"Invalid placement {placement_settings} from {valid_placement_keys}"
            )
        for k in placement_settings.keys():
            if k not in valid_placement_keys:
                raise ValueError(f"Invalid placement {k} from {valid_placement_keys}")

        x = placement_settings.get("x")
        y = placement_settings.get("y")
        dx = placement_settings.get("dx")
        dy = placement_settings.get("dy")
        port = placement_settings.get("port")
        rotation = placement_settings.get("rotation")
        mirror = placement_settings.get("mirror")

        if port:
            a = _get_anchor_point_from_name(ref, port)
            if a is None:
                raise ValueError(
                    f"Port {port} is neither a valid port on {ref.parent.name}"
                    " nor a recognized anchor keyword.\n"
                    "Valid ports: \n"
                    f"{list(ref.ports.keys())}. \n"
                    "Valid keywords: \n"
                    f"{valid_anchor_point_keywords}",
                )
            ref.x -= a[0]
            ref.y -= a[1]
        if x:
            if isinstance(x, str):
                if not len(x.split(",")) == 2:
                    raise ValueError(
                        f"You can define x as `x: instaceName,portName` got `x: {x!r}`"
                    )
                instance_name_ref, port_name = x.split(",")
                if instance_name_ref in all_remaining_insts:
                    place(
                        placements_conf,
                        connections_by_transformed_inst,
                        instances,
                        encountered_insts,
                        instance_name_ref,
                        all_remaining_insts,
                    )
                if instance_name_ref not in instances:
                    raise ValueError(
                        f"instance {instance_name_ref} not in {list(instances.keys())}."
                        f" You can define x as `x: instaceName,portName`, got x: {x!r}"
                    )
                if (
                    port_name not in instances[instance_name_ref].ports
                    and port_name not in valid_anchor_keywords
                ):
                    raise ValueError(
                        f"port = `{port_name}` not in {list(instances[instance_name_ref].ports.keys())}"
                        f" or in valid anchors {valid_anchor_keywords} for {instance_name_ref}, "
                        f"you can define x as `x: instaceName,portName`, got `x: {x!r}`"
                    )

                x = _get_anchor_value_from_name(
                    instances[instance_name_ref], port_name, "x"
                )
            ref.x += x
        if y:
            if isinstance(y, str):
                if not len(y.split(",")) == 2:
                    raise ValueError(
                        f"You can define y as `y: instaceName,portName` got `y: {y!r}`"
                    )
                instance_name_ref, port_name = y.split(",")
                if instance_name_ref in all_remaining_insts:
                    place(
                        placements_conf,
                        connections_by_transformed_inst,
                        instances,
                        encountered_insts,
                        instance_name_ref,
                        all_remaining_insts,
                    )
                if instance_name_ref not in instances:
                    raise ValueError(
                        f"{instance_name_ref} not in {list(instances.keys())}, "
                        f"you can define y as `y: instaceName,portName`, got `y: {y!r}`"
                    )
                if (
                    port_name not in instances[instance_name_ref].ports
                    and port_name not in valid_anchor_keywords
                ):
                    raise ValueError(
                        f"port = {port_name} not in {list(instances[instance_name_ref].ports.keys())} "
                        f"or in valid anchors {valid_anchor_keywords} for {instance_name_ref}, "
                        f"you can define y as `y: instaceName,portName`, got `y: {y!r}`"
                    )

                y = _get_anchor_value_from_name(
                    instances[instance_name_ref], port_name, "y"
                )
            ref.y += y
        if dx:
            ref.x += dx
        if dy:
            ref.y += dy
        if mirror:
            if mirror is True and port:
                ref.reflect_h(x0=_get_anchor_value_from_name(ref, port, "x"))
            elif mirror is True:
                if x:
                    ref.reflect_h(x0=x)
                else:
                    ref.reflect_h()
            elif mirror is False:
                pass
            elif isinstance(mirror, str):
                ref.reflect_h(port_name=mirror)
            elif isinstance(mirror, (int, float)):
                ref.reflect_h(x0=mirror)
            else:
                raise ValueError(
                    f"{mirror!r} can only be a port name {ref.ports.keys()}, "
                    "x value or True/False"
                )

        if rotation:
            if port:
                ref.rotate(rotation, center=_get_anchor_point_from_name(ref, port))
            else:
                x, y = ref.origin
                ref.rotate(rotation, center=(x, y))
                # ref.rotate(rotation, center=(ref.x, ref.y))

    if instance_name in connections_by_transformed_inst:
        conn_info = connections_by_transformed_inst[instance_name]
        instance_dst_name = conn_info["instance_dst_name"]
        if instance_dst_name in all_remaining_insts:
            place(
                placements_conf,
                connections_by_transformed_inst,
                instances,
                encountered_insts,
                instance_dst_name,
                all_remaining_insts,
            )

        make_connection(instances=instances, **conn_info)
        # placements_conf.pop(instance_name)


def transform_connections_dict(connections_conf: Dict[str, str]) -> Dict[str, Dict]:
    """Returns Dict with source_instance_name key and connection properties."""
    if not connections_conf:
        return {}
    attrs_by_src_inst = {}
    for port_src_string, port_dst_string in connections_conf.items():
        instance_src_name, port_src_name = port_src_string.split(",")
        instance_dst_name, port_dst_name = port_dst_string.split(",")
        attrs_by_src_inst[instance_src_name] = {
            "instance_src_name": instance_src_name,
            "port_src_name": port_src_name,
            "instance_dst_name": instance_dst_name,
            "port_dst_name": port_dst_name,
        }
    return attrs_by_src_inst


def make_connection(
    instance_src_name: str,
    port_src_name: str,
    instance_dst_name: str,
    port_dst_name: str,
    instances: Dict[str, ComponentReference],
) -> None:
    instance_src_name = instance_src_name.strip()
    instance_dst_name = instance_dst_name.strip()
    port_src_name = port_src_name.strip()
    port_dst_name = port_dst_name.strip()

    assert (
        instance_src_name in instances
    ), f"{instance_src_name} not in {list(instances.keys())}"
    assert (
        instance_dst_name in instances
    ), f"{instance_dst_name} not in {list(instances.keys())}"
    instance_src = instances[instance_src_name]
    instance_dst = instances[instance_dst_name]

    assert port_src_name in instance_src.ports, (
        f"{port_src_name} not in {list(instance_src.ports.keys())} for"
        f" {instance_src_name} "
    )
    assert port_dst_name in instance_dst.ports, (
        f"{port_dst_name} not in {list(instance_dst.ports.keys())} for"
        f" {instance_dst_name}"
    )
    port_dst = instance_dst.ports[port_dst_name]
    instance_src.connect(port=port_src_name, destination=port_dst)


sample_mmis = """

info:
    polarization: te
    wavelength: 1.55
    description: just a demo on adding metadata

instances:
    mmi_long:
      component: mmi1x2
      settings:
        width_mmi: 4.5
        length_mmi: 10
    mmi_short:
      component: mmi1x2
      settings:
        width_mmi: 4.5
        length_mmi: 5

placements:
    mmi_long:
        rotation: 180
        x: 100
        y: 100

routes:
    route_name1:
        links:
            mmi_short,o2: mmi_long,o1

ports:
    o1: mmi_short,o1
    o2: mmi_long,o2
    o3: mmi_long,o3
"""


def from_yaml(
    yaml_str: Union[str, pathlib.Path, IO[Any]],
    component_factory: ComponentFactoryDict = factory,
    routing_strategy: Dict[str, Callable] = routing_strategy_factories,
    cross_section_factory: Dict[str, CrossSectionFactory] = cross_section_factory,
    label_instance_function: Callable = add_instance_label,
    **kwargs,
) -> Component:
    """Returns a Component defined in YAML file or string.

    Args:
        yaml: YAML IO describing Component file or string (with newlines)
          (instances, placements, routes, ports, connections, names)
        component_factory: dict of functions {factory_name: factory_function}
        routing_strategy: for links
        label_instance_function: to label each instance
        kwargs: cache, prefix, autoname ... to pass to all factories

    Returns:
        Component

    .. code::

        valid properties:
        name: Optional Component name
        vars: Optional variables
        info: Optional component info
            description: just a demo
            polarization: TE
            ...
        instances:
            name:
                component: (ComponentFactory)
                settings (Optional)
                    length: 10
                    ...
        placements:
            x: Optional[float, str]  str can be instanceName,portName
            y: Optional[float, str]
            rotation: Optional[float]
            mirror: Optional[bool, float] float is x mirror axis
            port: Optional[str] port anchor
        connections (Optional): between instances
        ports (Optional): ports to expose
        routes (Optional): bundles of routes
            routeName:
            library: optical
            links:
                instance1,port1: instance2,port2


    .. code::

        vars:
            length_mmi: 5

        instances:
            mmi_bot:
              component: mmi1x2
              settings:
                width_mmi: 4.5
                length_mmi: 10
            mmi_top:
              component: mmi1x2
              settings:
                width_mmi: 4.5
                length_mmi: ${vars.length_mmi}

        placements:
            mmi_top:
                port: o1
                x: 0
                y: 0
            mmi_bot:
                port: o1
                x: mmi_top,o2
                y: mmi_top,o2
                dx: 30
                dy: -30
        routes:
            optical:
                library: optical
                links:
                    mmi_top,o3: mmi_bot,o1

    """
    yaml_str = (
        io.StringIO(yaml_str)
        if isinstance(yaml_str, str) and "\n" in yaml_str
        else yaml_str
    )

    conf = OmegaConf.load(yaml_str)  # nicer loader than conf = yaml.safe_load(yaml_str)
    for key in conf.keys():
        assert key in valid_top_level_keys, f"{key} not in {list(valid_top_level_keys)}"

    instances = {}
    routes = {}
    name = conf.get(
        "name",
        f"Unnamed_{hashlib.md5(json.dumps(OmegaConf.to_container(conf)).encode()).hexdigest()[:8]}",
    )
    if name in CACHE:
        return CACHE[name]
    else:
        c = Component(name)
        CACHE[name] = c
    placements_conf = conf.get("placements")
    routes_conf = conf.get("routes")
    ports_conf = conf.get("ports")
    connections_conf = conf.get("connections")
    instances_dict = conf["instances"]
    c.info = conf.get("info", omegaconf.DictConfig({}))

    for instance_name in instances_dict:
        instance_conf = instances_dict[instance_name]
        component_type = instance_conf["component"]
        assert (
            component_type in component_factory
        ), f"{component_type} not in {list(component_factory.keys())}"

        settings = instance_conf.get("settings", {})
        settings = OmegaConf.to_container(settings, resolve=True) if settings else {}
        settings.update(**kwargs)

        if "cross_section" in settings:
            name_or_dict = settings["cross_section"]
            if isinstance(name_or_dict, str):
                cross_section = cross_section_factory[name_or_dict]
            elif isinstance(name_or_dict, dict):
                name = name_or_dict.pop("function")
                cross_section = functools.partial(
                    cross_section_factory[name], **name_or_dict
                )
            else:
                raise ValueError(f"invalid type for cross_section={name_or_dict!r}")
            settings["cross_section"] = cross_section

        ci = component_factory[component_type](**settings)
        ref = c << ci
        instances[instance_name] = ref

    placements_conf = dict() if placements_conf is None else placements_conf

    connections_by_transformed_inst = transform_connections_dict(connections_conf)
    components_to_place = set(placements_conf.keys())
    components_with_placement_conflicts = components_to_place.intersection(
        connections_by_transformed_inst.keys()
    )
    for instance_name in components_with_placement_conflicts:
        placement_settings = placements_conf[instance_name]
        if "x" in placement_settings or "y" in placement_settings:
            warnings.warn(
                f"YAML defined: ({', '.join(components_with_placement_conflicts)}) "
                + "with both connection and placement. Please use one or the other.",
            )

    all_remaining_insts = list(
        set(placements_conf.keys()).union(set(connections_by_transformed_inst.keys()))
    )

    while all_remaining_insts:
        place(
            placements_conf=placements_conf,
            connections_by_transformed_inst=connections_by_transformed_inst,
            instances=instances,
            encountered_insts=list(),
            all_remaining_insts=all_remaining_insts,
        )

    for instance_name in instances_dict:
        label_instance_function(
            component=c, instance_name=instance_name, reference=instances[instance_name]
        )

    if routes_conf:
        for route_alias in routes_conf:
            route_names = []
            ports1 = []
            ports2 = []
            routes_dict = routes_conf[route_alias]
            for key in routes_dict.keys():
                if key not in valid_route_keys:
                    raise ValueError(
                        f"{route_alias!r} key={key!r} not in {valid_route_keys}"
                    )

            settings = routes_dict.pop("settings", {})
            settings = (
                OmegaConf.to_container(settings, resolve=True) if settings else {}
            )
            if "cross_section" in settings:
                name_or_dict = settings["cross_section"]
                if isinstance(name_or_dict, str):
                    cross_section = cross_section_factory[name_or_dict]
                elif isinstance(name_or_dict, dict):
                    name = name_or_dict.pop("function")
                    cross_section = functools.partial(
                        cross_section_factory[name], **name_or_dict
                    )
                else:
                    raise ValueError(f"invalid type for cross_section={name_or_dict}")
                settings["cross_section"] = cross_section
            routing_strategy_name = routes_dict.pop("routing_strategy", "get_bundle")
            if routing_strategy_name not in routing_strategy:
                raise ValueError(
                    f"function {routing_strategy_name!r} not in routing_strategy {list(routing_strategy.keys())}"
                )

            if "links" not in routes_dict:
                raise ValueError(
                    f"You need to define links for the {route_alias!r} route"
                )
            links_dict = routes_dict["links"]

            for port_src_string, port_dst_string in links_dict.items():

                if ":" in port_src_string:
                    src, src0, src1 = [s.strip() for s in port_src_string.split(":")]
                    dst, dst0, dst1 = [s.strip() for s in port_dst_string.split(":")]
                    instance_src_name, port_src_name = [
                        s.strip() for s in src.split(",")
                    ]
                    instance_dst_name, port_dst_name = [
                        s.strip() for s in dst.split(",")
                    ]

                    src0 = int(src0)
                    src1 = int(src1)
                    dst0 = int(dst0)
                    dst1 = int(dst1)

                    if src1 > src0:
                        ports1names = [
                            f"{port_src_name}{i}" for i in range(src0, src1 + 1, 1)
                        ]
                    else:
                        ports1names = [
                            f"{port_src_name}{i}" for i in range(src0, src1 - 1, -1)
                        ]

                    if dst1 > dst0:
                        ports2names = [
                            f"{port_dst_name}{i}" for i in range(dst0, dst1 + 1, 1)
                        ]
                    else:
                        ports2names = [
                            f"{port_dst_name}{i}" for i in range(dst0, dst1 - 1, -1)
                        ]

                    assert len(ports1names) == len(ports2names)
                    route_names += [
                        f"{instance_src_name},{i}:{instance_dst_name},{j}"
                        for i, j in zip(ports1names, ports2names)
                    ]

                    instance_src = instances[instance_src_name]
                    instance_dst = instances[instance_dst_name]

                    for port_src_name in ports1names:
                        assert port_src_name in instance_src.ports, (
                            f"{port_src_name} not in {list(instance_src.ports.keys())}"
                            f"for {instance_src_name} "
                        )
                        ports1.append(instance_src.ports[port_src_name])

                    for port_dst_name in ports2names:
                        assert port_dst_name in instance_dst.ports, (
                            f"{port_dst_name} not in {list(instance_dst.ports.keys())}"
                            f"for {instance_dst_name}"
                        )
                        ports2.append(instance_dst.ports[port_dst_name])

                    # print(ports1)
                    # print(ports2)
                    # print(route_names)

                else:
                    instance_src_name, port_src_name = port_src_string.split(",")
                    instance_dst_name, port_dst_name = port_dst_string.split(",")

                    instance_src_name = instance_src_name.strip()
                    instance_dst_name = instance_dst_name.strip()
                    port_src_name = port_src_name.strip()
                    port_dst_name = port_dst_name.strip()

                    assert (
                        instance_src_name in instances
                    ), f"{instance_src_name} not in {list(instances.keys())}"
                    assert (
                        instance_dst_name in instances
                    ), f"{instance_dst_name} not in {list(instances.keys())}"

                    instance_src = instances[instance_src_name]
                    instance_dst = instances[instance_dst_name]

                    assert port_src_name in instance_src.ports, (
                        f"{port_src_name} not in {list(instance_src.ports.keys())} for"
                        f" {instance_src_name} "
                    )
                    assert port_dst_name in instance_dst.ports, (
                        f"{port_dst_name} not in {list(instance_dst.ports.keys())} for"
                        f" {instance_dst_name}"
                    )

                    ports1.append(instance_src.ports[port_src_name])
                    ports2.append(instance_dst.ports[port_dst_name])
                    route_name = f"{port_src_string}:{port_dst_string}"
                    route_names.append(route_name)

            routing_function = routing_strategy[routing_strategy_name]
            route_or_route_list = routing_function(
                ports1=ports1,
                ports2=ports2,
                **settings,
            )

            # FIXME, be more consistent
            if isinstance(route_or_route_list, list):
                for route_name, route_dict in zip(route_names, route_or_route_list):
                    c.add(route_dict.references)
                    routes[route_name] = route_dict.length
            elif isinstance(route_or_route_list, Route):
                c.add(route_or_route_list.references)
                routes[route_name] = route_or_route_list.length
            else:
                raise ValueError(f"{route_or_route_list} needs to be a Route or a list")

    if ports_conf:
        assert hasattr(ports_conf, "items"), f"{ports_conf} needs to be a dict"
        for port_name, instance_comma_port in ports_conf.items():
            if "," in instance_comma_port:
                instance_name, instance_port_name = instance_comma_port.split(",")
                instance_name = instance_name.strip()
                instance_port_name = instance_port_name.strip()
                assert (
                    instance_name in instances
                ), f"{instance_name} not in {list(instances.keys())}"
                instance = instances[instance_name]
                assert instance_port_name in instance.ports, (
                    f"{instance_port_name} not in {list(instance.ports.keys())} for"
                    f" {instance_name} "
                )
                c.add_port(port_name, port=instance.ports[instance_port_name])
            else:
                c.add_port(**instance_comma_port)
    c.routes = routes
    c.instances = instances
    return c


if __name__ == "__main__":
    # for k in factory.keys():
    #     print(k)
    # print(c.settings["info"])

    # from gdsfactory.tests.test_component_from_yaml import yaml_anchor
    # c = from_yaml(yaml_anchor)
    c = from_yaml(sample_mmis)
    c.show()

    # c = test_connections_regex()
    # c = from_yaml(sample_regex_connections)
    # c = from_yaml(sample_regex_connections_backwards)
    # c = test_docstring_sample()
    # c = test_connections()
    # c = from_yaml(sample_mirror_simple)
    # c = test_connections_2x2()
    # c = test_connections_different_factory()
    # test_connections_different_link_factory()
    # test_connections_waypoints()
    # test_mirror()
    # c = from_yaml(sample_different_link_factory)
    # c = test_mirror()
    # c = from_yaml(sample_waypoints)
    # c = from_yaml(sample_2x2_connections)
    # c = from_yaml(sample_mmis)
    # c = from_yaml(sample_connections)
    # assert len(c.get_dependencies()) == 3
    # test_component_from_yaml()
    # test_component_from_yaml_with_routing()
    # print(c.ports)
    # c = gf.routing.add_fiber_array(c)



import itertools as it

import jsondiff
import pytest
from omegaconf import OmegaConf
from pytest_regressions.data_regression import DataRegressionFixture

import gdsfactory as gf
from gdsfactory import components

factory = {
    i: getattr(components, i)
    for i in dir(components)
    if not i.startswith("_") and callable(getattr(components, i))
}

circuit_names = {
    "mzi",
    "ring_single",
    "ring_single_array",
    "ring_double",
    "mzit_lattice",
    "mzit",
    "component_lattice",
}


circuit_names_test = circuit_names - {
    "component_lattice",
    "mzi",
}  # set of component names


@pytest.mark.parametrize(
    "component_type,full_settings", it.product(circuit_names_test, [True, False])
)
def test_netlists(
    component_type: str,
    full_settings: bool,
    data_regression: DataRegressionFixture,
    check: bool = True,
    component_factory=factory,
) -> None:
    """Write netlists for hierarchical circuits.
    Checks that both netlists are the same
    jsondiff does a hierarchical diff

    Component -> netlist -> Component -> netlist
    """
    c = component_factory[component_type]()
    n = c.get_netlist(full_settings=full_settings)
    if check:
        data_regression.check(OmegaConf.to_container(n))

    yaml_str = OmegaConf.to_yaml(n, sort_keys=True)
    # print(yaml_str)
    c2 = gf.read.from_yaml(yaml_str, component_factory=component_factory)
    n2 = c2.get_netlist(full_settings=full_settings)

    d = jsondiff.diff(n, n2)
    # print(yaml_str)
    # print(d)
    # yaml_str2 = OmegaConf.to_yaml(n2, sort_keys=True)
    # print(yaml_str2)
    assert len(d) == 0, d


def demo_netlist(component_type):
    c1 = factory[component_type]()
    n = c1.get_netlist()
    yaml_str = OmegaConf.to_yaml(n, sort_keys=True)
    c2 = gf.read.from_yaml(yaml_str)
    gf.show(c2)


if __name__ == "__main__":
    # c = factory["ring_double"]()
    # n = c.get_netlist()

    # c = factory["mzi"]()
    # c = factory["ring_double"]()

    # gf.clear_connections()
    # print(n.connections)
    # n = c.get_netlist_yaml()
    # print(n)
    # c.show()

    # c = factory["ring_single"]()
    # n.pop("connections")
    # n.pop("placements")

    # component_type = "mzi"
    # component_type = "mzit"
    # component_type = "ring_double"
    component_type = "ring_single"
    c1 = factory[component_type]()
    full_settings = True
    n = c1.get_netlist(full_settings=full_settings)
    yaml_str = OmegaConf.to_yaml(n, sort_keys=True)
    print(yaml_str)
    c2 = gf.read.from_yaml(yaml_str)
    n2 = c2.get_netlist(full_settings=full_settings)
    d = jsondiff.diff(n, n2)
    print(d)
    gf.show(c2)

    # test_netlists(
    #     component_type="mzi",
    #     # component_type="ring_single",
    #     full_settings=True,
    #     data_regression=None,
    #     check=False,
    # )



import pathlib
from typing import Union

from gdsfactory import klive
from gdsfactory.cell import clear_cache as clear_cache_function
from gdsfactory.component import Component
from gdsfactory.config import logger


def show(
    component: Union[Component, str, pathlib.Path], clear_cache: bool = True, **kwargs
) -> None:
    """Write GDS and show Component in klayout

    Args:
        component
        clear_cache: clear_cache after showing the component

    Keyword Args:
        gdspath: GDS file path to write to.
        gdsdir: directory for the GDS file. Defaults to /tmp/
        unit: unit size for objects in library. 1um by default.
        precision: for object dimensions in the library (m). 1nm by default.
        timestamp: Defaults to 2019-10-25. If None uses current time.

    """
    if isinstance(component, pathlib.Path):
        component = str(component)
        return klive.show(component)
    elif isinstance(component, str):
        return klive.show(component)
    elif component is None:
        raise ValueError(
            "Component is None, make sure that your function returns the component"
        )

    elif isinstance(component, Component):
        gdspath = component.write_gds(logging=False, **kwargs)
        klive.show(gdspath)
        logger.info(f"Klayout show {component!r}")
    else:
        raise ValueError(
            f"Component is {type(component)}, make sure pass a Component or a path"
        )
    if clear_cache:
        clear_cache_function()



from gdsfactory.sweep.read_sweep import get_settings_list, read_sweep, test_read_sweep
from gdsfactory.sweep.write_sweep import (
    get_markdown_table,
    write_sweep,
    write_sweep_metadata,
)
from gdsfactory.sweep.write_sweep_from_yaml import write_sweep_from_yaml
from gdsfactory.sweep.write_sweeps import write_sweeps

__all__ = [
    "get_markdown_table",
    "get_settings_list",
    "read_sweep",
    "test_read_sweep",
    "write_sweep",
    "write_sweep_from_yaml",
    "write_sweep_metadata",
    "write_sweeps",
]



import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.wire import wire_corner, wire_straight


@gf.cell
def wire_sbend(dx: float = 20.0, dy: float = 10.0, **kwargs) -> Component:
    """Sbend corner with manhattan wires

    Args:
        dx: length
        dy: height
        **kwargs: cross_section settings
    """
    sx = wire_straight(length=dx / 2, **kwargs)
    sy = wire_straight(length=dy, **kwargs)
    bc = wire_corner(**kwargs)

    symbol_to_component = {
        "-": (sx, "e1", "e2"),
        "|": (sy, "e1", "e2"),
        "b": (bc, "e2", "e1"),
        "B": (bc, "e1", "e2"),
    }

    sequence = "-B|b-"
    c = gf.components.component_sequence(
        sequence=sequence, symbol_to_component=symbol_to_component
    )
    c.auto_rename_ports()
    return c


if __name__ == "__main__":

    c = wire_sbend(width=5)
    c.show(show_ports=True)
    c.pprint_ports()



from typing import Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
from pandas import DataFrame

import gdsfactory as gf


def plot_sparameters(
    df: DataFrame,
    logscale: bool = True,
    keys: Optional[Tuple[str, ...]] = None,
    **sim_settings,
):
    """Plots Sparameters from a pandas DataFrame.

    Args:
        df: Sparameters pandas DataFrame
        logscale: plots 20*log10(S)
        keys: list of keys to plot, plots all by default.

    Keyword Args:
        sim_settings: simulation settings for the write_sparameters_function

    """

    w = df["wavelengths"] * 1e3
    keys = keys or [
        key for key in df.keys() if key.lower().startswith("s") and key.endswith("m")
    ]

    for key in keys:
        if key in df:
            y = df[key]
            y = 20 * np.log10(y) if logscale else y
            plt.plot(w, y, label=key[:-1])
        else:
            raise ValueError(f"{key} not in {df.keys()}")
    plt.legend()
    plt.xlabel("wavelength (nm)")
    plt.ylabel("|S| (dB)") if logscale else plt.ylabel("|S|")


def plot_imbalance2x2(df: DataFrame, port1: str = "s13m", port2: str = "s14m") -> None:
    """Plots imbalance in % for 2x2 coupler"""
    y1 = df[port1].values
    y2 = df[port2].values
    imbalance = y1 / y2
    plt.plot(df.wavelength_nm, 100 * abs(imbalance))
    plt.xlabel("wavelength (nm)")
    plt.ylabel("imbalance (%)")
    plt.grid()


def plot_loss2x2(df: DataFrame, port1: str = "s13m", port2: str = "s14m") -> None:
    """Plots imbalance in % for 2x2 coupler"""
    y1 = df[port1].values
    y2 = df[port2].values
    plt.plot(df.wavelength_nm, abs(10 * np.log10(y1 ** 2 + y2 ** 2)))
    plt.xlabel("wavelength")
    plt.ylabel("excess loss (dB)")


plot_loss1x2 = gf.partial(plot_loss2x2, port1="S13m", port2="S12m")
plot_imbalance1x2 = gf.partial(plot_imbalance2x2, port1="S13m", port2="S12m")


if __name__ == "__main__":
    import gdsfactory.simulation as sim

    df = sim.get_sparameters_data_lumerical(component=gf.c.mmi1x2)
    plot_sparameters(df, logscale=True)
    plt.show()



from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.cross import cross
from gdsfactory.types import ComponentFactory, Layers


@cell
def copy_layers(
    factory: ComponentFactory = cross, layers: Layers = ((1, 0), (2, 0)), **kwargs
) -> Component:
    """Returns a component with the geometry copied in different layers.

    Args:
        factory: component factory / function
        layers: iterable of layers
        kwargs: keyword arguments
    """
    c = Component()
    for layer in layers:
        ci = factory(layer=layer, **kwargs)
        c << ci

    c.copy_child_info(ci)
    return c


if __name__ == "__main__":
    import gdsfactory as gf

    c = copy_layers(gf.components.rectangle)
    c.show()



"""wires for electrical manhattan routes
"""

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.straight import straight
from gdsfactory.cross_section import metal3
from gdsfactory.types import CrossSectionFactory

wire_straight = gf.partial(straight, with_cladding_box=False, cross_section=metal3)


@gf.cell
def wire_corner(cross_section: CrossSectionFactory = metal3, **kwargs) -> Component:
    """90 degrees electrical corner

    Args:
        waveguide:
        kwargs: cross_section settings

    """
    x = cross_section(**kwargs)
    layer = x.info["layer"]
    width = x.info["width"]

    c = Component()
    a = width / 2
    xpts = [-a, a, a, -a]
    ypts = [-a, -a, a, a]

    c.add_polygon([xpts, ypts], layer=layer)
    c.add_port(
        name="e1",
        midpoint=(-a, 0),
        width=width,
        orientation=180,
        layer=layer,
        port_type="electrical",
    )
    c.add_port(
        name="e2",
        midpoint=(0, a),
        width=width,
        orientation=90,
        layer=layer,
        port_type="electrical",
    )
    c.info.length = width
    return c


if __name__ == "__main__":

    # c = wire_straight()
    c = wire_corner()
    c.show(show_ports=True)
    c.pprint_ports()



"""Write Sparameters with for different components."""
from typing import Optional

from gdsfactory.simulation.lumerical.write_sparameters_lumerical import (
    write_sparameters_lumerical,
)
from gdsfactory.types import ComponentFactoryDict


def write_sparameters_components_lumerical(
    factory: ComponentFactoryDict,
    run: bool = False,
    session: Optional[object] = None,
    **kwargs,
) -> None:
    """Writes component Sparameters using Lumerical FDTD.

    Args:
        factory: dict of component functions to simulate.
        run: if False, does not run and prompts you to review each simulation
        session: lumapi.FDTD() Lumerical FDTD session

    Keyword Args:
        simulation settings
    """
    import lumapi

    session = session or lumapi.FDTD()
    need_review = []

    for component_name in factory.keys():
        component = factory[component_name]()
        write_sparameters_lumerical(component, run=run, session=session, **kwargs)
        if not run:
            response = input(
                f"does the simulation for {component_name} look good? (y/n)"
            )
            if response.upper()[0] == "N":
                need_review.append(component_name)


if __name__ == "__main__":
    from gdsfactory.components import _factory_passives

    write_sparameters_components_lumerical(factory=_factory_passives)



import pathlib

import pytest
from pytest_regressions.data_regression import DataRegressionFixture

from gdsfactory.components import factory
from gdsfactory.difftest import difftest

skip_test = {
    "version_stamp",
    "extend_ports_list",
    "extend_port",
    "component_sequence",
    "mzi_arm",
}

components_to_test = set(factory.keys()) - skip_test
dirpath = pathlib.Path(__file__).absolute().with_suffix(".gds")


@pytest.fixture(params=components_to_test, scope="function")
def component_name(request) -> str:
    return request.param


def test_gds(component_name: str) -> None:
    """Avoid regressions in GDS geometry shapes and layers."""
    component = factory[component_name]()
    difftest(component, test_name=component_name, dirpath=dirpath)


def test_settings(component_name: str, data_regression: DataRegressionFixture) -> None:
    """Avoid regressions when exporting settings."""
    component = factory[component_name]()
    data_regression.check(component.to_dict())


def test_assert_ports_on_grid(component_name: str):
    """Ensure ports are on grid."""
    component = factory[component_name]()
    component.assert_ports_on_grid()


if __name__ == "__main__":
    # c = gf.components.coupler(length=1.0 + 1e-4)
    # c = gf.components.coupler()
    # c.assert_ports_on_grid()
    # print(c.ports)
    print(dirpath)



import datetime
import pathlib
from typing import Optional

import gdspy

from gdsfactory.component import _timestamp2019
from gdsfactory.config import CONFIG, logger
from gdsfactory.name import clean_name
from gdsfactory.types import PathType

script_prefix = """
from pathlib import PosixPath
from functools import partial
import gdsfactory as gf

add_ports_optical = gf.partial(gf.add_ports.add_ports_from_markers_inside, pin_layer=(1, 0), port_layer=(2, 0))
add_ports_electrical = gf.partial(gf.add_ports.add_ports_from_markers_inside, pin_layer=(41, 0), port_layer=(1, 0))
add_ports = gf.compose(add_ports_optical, add_ports_electrical)

"""


def get_import_gds_script(dirpath: PathType) -> str:
    """Returns import_gds script from a directory with all the GDS files."""
    dirpath = pathlib.Path(dirpath)
    script = [script_prefix]
    script += [f"gdsdir = {dirpath.absolute()!r}\n"]
    script += [
        "import_gds = partial(gf.import_gds, gdsdir=gdsdir, decorator=add_ports)\n"
    ]

    cells = [
        f"{clean_name(cell.stem)} = partial(import_gds, "
        f"{cell.stem + cell.suffix!r})"
        for cell in dirpath.glob("*.gds")
    ]
    script += sorted(cells)
    return "\n".join(script)


def write_cells_recursively(
    cell: gdspy.Cell,
    unit: float = 1e-6,
    precision: float = 1e-9,
    timestamp: Optional[datetime.datetime] = _timestamp2019,
    dirpath: pathlib.Path = Optional[pathlib.Path],
):
    """Write gdspy cells recursively

    Args:
        cell: gdspy cell
        unit: unit size for objects in library. 1um by default.
        precision: for object dimensions in the library (m). 1nm by default.
        timestamp: Defaults to 2019-10-25. If None uses current time.
        dirpath: directory for the GDS file
    """
    dirpath = dirpath or pathlib.Path.cwd()

    for cell in cell.get_dependencies():
        gdspath = f"{pathlib.Path(dirpath)/cell.name}.gds"
        lib = gdspy.GdsLibrary(unit=unit, precision=precision)
        lib.write_gds(gdspath, cells=[cell], timestamp=timestamp)
        logger.info(f"Write GDS to {gdspath}")

        if cell.get_dependencies():
            write_cells_recursively(
                cell=cell,
                unit=unit,
                precision=precision,
                timestamp=timestamp,
                dirpath=dirpath,
            )


def write_cells(
    gdspath: Optional[PathType] = None,
    cell: Optional[gdspy.Cell] = None,
    dirpath: Optional[PathType] = None,
    unit: float = 1e-6,
    precision: float = 1e-9,
    timestamp: Optional[datetime.datetime] = _timestamp2019,
    recursively: bool = True,
) -> None:
    """Writes cells into separate GDS files.

    Args:
        gdspath: GDS file. You need to define either gdspath or cell.
        cell: gdspy cell. You need to define either gdspath or cell.
        unit: unit size for objects in library. 1um by default.
        precision: for object dimensions in the library (m). 1nm by default.
        timestamp: Defaults to 2019-10-25. If None uses current time.
        dirpath: directory for the GDS file. Defaults to current working directory.
        recursively: writes all cells recursively. If False writes only top cells.
    """
    if cell is None and gdspath is None:
        raise ValueError("You need to specify component or gdspath")

    gdsii_lib = gdspy.GdsLibrary()
    gdsii_lib.read_gds(gdspath)
    top_level_cells = gdsii_lib.top_level()

    dirpath = dirpath or pathlib.Path.cwd()
    dirpath = pathlib.Path(dirpath)
    dirpath.mkdir(exist_ok=True, parents=True)

    for cell in top_level_cells:
        gdspath = f"{pathlib.Path(dirpath)/cell.name}.gds"
        lib = gdspy.GdsLibrary(unit=unit, precision=precision)
        lib.write_gds(gdspath, cells=[cell], timestamp=timestamp)
        logger.info(f"Write GDS to {gdspath}")

        if recursively:
            write_cells_recursively(
                cell=cell,
                unit=unit,
                precision=precision,
                timestamp=timestamp,
                dirpath=dirpath,
            )


if __name__ == "__main__":
    import gdsfactory as gf

    gdspath = CONFIG["gdsdir"] / "mzi2x2.gds"
    gf.show(gdspath)

    write_cells(gdspath=gdspath, dirpath="extra/gds")



from gdsfactory.simulation.lumerical.read import read_sparameters_lumerical
from gdsfactory.simulation.lumerical.write_sparameters_components_lumerical import (
    write_sparameters_components_lumerical,
)
from gdsfactory.simulation.lumerical.write_sparameters_lumerical import (
    write_sparameters_lumerical,
)

__all__ = [
    "read_sparameters_lumerical",
    "write_sparameters_components_lumerical",
    "write_sparameters_lumerical",
]



from functools import partial
from typing import Optional, Tuple

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.compass import compass
from gdsfactory.tech import LAYER
from gdsfactory.types import ComponentOrFactory, Layer


@cell
def pad(
    size: Tuple[float, float] = (100.0, 100.0),
    layer: Layer = LAYER.M3,
    layers_cladding: Optional[Tuple[Layer, ...]] = None,
    cladding_offsets: Optional[Tuple[float, ...]] = None,
    port_inclusion: float = 0,
) -> Component:
    """Rectangular pad with 4 ports (1, 2, 3, 4)

    Args:
        size:
        layer: pad layer
        layers_cladding:
        cladding_offsets:
        port_inclusion: from edge
    """
    c = Component()
    rect = compass(size=size, layer=layer, port_inclusion=port_inclusion)
    c_ref = c.add_ref(rect)
    c.add_ports(c_ref.ports)
    c.info.size = (float(size[0]), float(size[1]))
    c.info.layer = layer

    if layers_cladding and cladding_offsets:
        for layer, cladding_offset in zip(layers_cladding, cladding_offsets):
            c.add_ref(
                compass(
                    size=(size[0] + 2 * cladding_offset, size[1] + 2 * cladding_offset),
                    layer=layer,
                )
            )

    c.add_port(name="pad", port_type="vertical_dc", layer=layer, orientation=0)
    return c


@cell
def pad_array(
    pad: ComponentOrFactory = pad,
    spacing: Tuple[float, float] = (150.0, 150.0),
    columns: int = 6,
    rows: int = 1,
    orientation: int = 270,
) -> Component:
    """Returns 2D array of pads

    Args:
        pad: pad element
        spacing: x, y pitch
        columns:
        rows:
        orientation: port orientation in deg
    """
    c = Component()
    pad = pad() if callable(pad) else pad
    size = pad.info.full.size
    c.info.size = size

    c.add_array(pad, columns=columns, rows=rows, spacing=spacing)
    width = size[0] if orientation in [90, 270] else size[1]

    for col in range(columns):
        for row in range(rows):
            c.add_port(
                name=f"e{row+1}{col+1}",
                midpoint=(col * spacing[0], row * spacing[1]),
                width=width,
                orientation=orientation,
                port_type="electrical",
                layer=pad.info["layer"],
            )
    return c


pad_array90 = partial(pad_array, orientation=90)
pad_array270 = partial(pad_array, orientation=270)


if __name__ == "__main__":
    c = pad()
    # c = pad(layer_to_inclusion={(3, 0): 10})
    # print(c.ports)
    # c = pad(width=10, height=10)
    # print(c.ports.keys())
    # c = pad_array90()
    # c = pad_array270()
    # c.pprint_ports()
    # c = pad_array_2d(cols=2, rows=3, port_names=("e2",))
    # c = pad_array(columns=2, rows=2, orientation=270)
    # c.auto_rename_ports()
    c.show()



import pytest
from pytest_regressions.data_regression import DataRegressionFixture

import gdsfactory as gf
from gdsfactory import Port
from gdsfactory.component import Component
from gdsfactory.difftest import difftest
from gdsfactory.routing.get_bundle import get_bundle


def test_get_bundle(data_regression: DataRegressionFixture, check: bool = True):
    xs_top = [-100, -90, -80, 0, 10, 20, 40, 50, 80, 90, 100, 105, 110, 115]
    pitch = 127.0
    N = len(xs_top)
    xs_bottom = [(i - N / 2) * pitch for i in range(N)]

    top_ports = [Port(f"top_{i}", (xs_top[i], 0), 0.5, 270) for i in range(N)]

    bot_ports = [Port(f"bot_{i}", (xs_bottom[i], -400), 0.5, 90) for i in range(N)]

    c = gf.Component("test_get_bundle")
    routes = get_bundle(
        top_ports, bot_ports, start_straight_length=5, end_straight_length=10
    )
    lengths = {}
    for i, route in enumerate(routes):
        c.add(route.references)
        lengths[i] = route.length

    if check:
        data_regression.check(lengths)
        difftest(c)
    return c


@pytest.mark.parametrize("config", ["A", "C"])
def test_connect_corner(
    config: str, data_regression: DataRegressionFixture, check: bool = True, N=6
) -> Component:
    d = 10.0
    sep = 5.0
    c = Component(name=f"test_connect_corner_{config}")

    if config in ["A", "B"]:
        a = 100.0
        ports_A_TR = [
            Port("A_TR_{}".format(i), (d, a / 2 + i * sep), 0.5, 0) for i in range(N)
        ]
        ports_A_TL = [
            Port("A_TL_{}".format(i), (-d, a / 2 + i * sep), 0.5, 180) for i in range(N)
        ]
        ports_A_BR = [
            Port("A_BR_{}".format(i), (d, -a / 2 - i * sep), 0.5, 0) for i in range(N)
        ]
        ports_A_BL = [
            Port("A_BL_{}".format(i), (-d, -a / 2 - i * sep), 0.5, 180)
            for i in range(N)
        ]

        ports_A = [ports_A_TR, ports_A_TL, ports_A_BR, ports_A_BL]

        ports_B_TR = [
            Port("B_TR_{}".format(i), (a / 2 + i * sep, d), 0.5, 90) for i in range(N)
        ]
        ports_B_TL = [
            Port("B_TL_{}".format(i), (-a / 2 - i * sep, d), 0.5, 90) for i in range(N)
        ]
        ports_B_BR = [
            Port("B_BR_{}".format(i), (a / 2 + i * sep, -d), 0.5, 270) for i in range(N)
        ]
        ports_B_BL = [
            Port("B_BL_{}".format(i), (-a / 2 - i * sep, -d), 0.5, 270)
            for i in range(N)
        ]

        ports_B = [ports_B_TR, ports_B_TL, ports_B_BR, ports_B_BL]

    elif config in ["C", "D"]:
        a = N * sep + 2 * d
        ports_A_TR = [
            Port("A_TR_{}".format(i), (a, d + i * sep), 0.5, 0) for i in range(N)
        ]
        ports_A_TL = [
            Port("A_TL_{}".format(i), (-a, d + i * sep), 0.5, 180) for i in range(N)
        ]
        ports_A_BR = [
            Port("A_BR_{}".format(i), (a, -d - i * sep), 0.5, 0) for i in range(N)
        ]
        ports_A_BL = [
            Port("A_BL_{}".format(i), (-a, -d - i * sep), 0.5, 180) for i in range(N)
        ]

        ports_A = [ports_A_TR, ports_A_TL, ports_A_BR, ports_A_BL]

        ports_B_TR = [
            Port("B_TR_{}".format(i), (d + i * sep, a), 0.5, 90) for i in range(N)
        ]
        ports_B_TL = [
            Port("B_TL_{}".format(i), (-d - i * sep, a), 0.5, 90) for i in range(N)
        ]
        ports_B_BR = [
            Port("B_BR_{}".format(i), (d + i * sep, -a), 0.5, 270) for i in range(N)
        ]
        ports_B_BL = [
            Port("B_BL_{}".format(i), (-d - i * sep, -a), 0.5, 270) for i in range(N)
        ]

        ports_B = [ports_B_TR, ports_B_TL, ports_B_BR, ports_B_BL]

    lengths = {}
    i = 0
    if config in ["A", "C"]:
        for ports1, ports2 in zip(ports_A, ports_B):
            routes = get_bundle(ports1, ports2)
            for route in routes:
                c.add(route.references)
                lengths[i] = route.length
                i += 1

    elif config in ["B", "D"]:
        for ports1, ports2 in zip(ports_A, ports_B):
            routes = get_bundle(ports2, ports1)
            for route in routes:
                c.add(route.references)
                lengths[i] = route.length
                i += 1

    if check:
        data_regression.check(lengths)
        difftest(c)
    return c


def test_get_bundle_udirect(
    data_regression: DataRegressionFixture, check: bool = True, dy=200, angle=270
):
    xs1 = [-100, -90, -80, -55, -35, 24, 0] + [200, 210, 240]
    axis = "X" if angle in [0, 180] else "Y"

    pitch = 10.0
    N = len(xs1)
    xs2 = [70 + i * pitch for i in range(N)]

    if axis == "X":
        ports1 = [Port(f"top_{i}", (0, xs1[i]), 0.5, angle) for i in range(N)]
        ports2 = [Port(f"bot_{i}", (dy, xs2[i]), 0.5, angle) for i in range(N)]

    else:
        ports1 = [Port(f"top_{i}", (xs1[i], 0), 0.5, angle) for i in range(N)]
        ports2 = [Port(f"bot_{i}", (xs2[i], dy), 0.5, angle) for i in range(N)]

    c = gf.Component(name="test_get_bundle_udirect")
    routes = get_bundle(
        ports1, ports2, bend=gf.components.bend_circular, end_straight_length=30
    )
    lengths = {}
    for i, route in enumerate(routes):
        c.add(route.references)
        lengths[i] = route.length

    if check:
        data_regression.check(lengths)
        difftest(c)

    return c


@pytest.mark.parametrize("angle", [0, 90, 180, 270])
def test_get_bundle_u_indirect(
    data_regression: DataRegressionFixture, angle, check: bool = True, dy=-200
):

    xs1 = [-100, -90, -80, -55, -35] + [200, 210, 240]

    axis = "X" if angle in [0, 180] else "Y"

    pitch = 10.0
    N = len(xs1)
    xs2 = [50 + i * pitch for i in range(N)]

    a1 = angle
    a2 = a1 + 180

    if axis == "X":
        ports1 = [Port("top_{}".format(i), (0, xs1[i]), 0.5, a1) for i in range(N)]

        ports2 = [Port("bot_{}".format(i), (dy, xs2[i]), 0.5, a2) for i in range(N)]

    else:
        ports1 = [Port("top_{}".format(i), (xs1[i], 0), 0.5, a1) for i in range(N)]

        ports2 = [Port("bot_{}".format(i), (xs2[i], dy), 0.5, a2) for i in range(N)]

    c = gf.Component(f"test_get_bundle_u_indirect_{angle}_{dy}")

    routes = get_bundle(
        ports1,
        ports2,
        bend=gf.components.bend_circular,
        end_straight_length=15,
        start_straight_length=5,
    )
    lengths = {}
    for i, route in enumerate(routes):
        c.add(route.references)
        lengths[i] = route.length

    if check:
        data_regression.check(lengths)
        difftest(c)

    return c


def test_facing_ports(
    data_regression: DataRegressionFixture,
    check: bool = True,
):

    dy = 200.0
    xs1 = [-500, -300, -100, -90, -80, -55, -35, 200, 210, 240, 500, 650]

    pitch = 10.0
    N = len(xs1)
    xs2 = [-20 + i * pitch for i in range(N // 2)]
    xs2 += [400 + i * pitch for i in range(N // 2)]

    a1 = 90
    a2 = a1 + 180

    ports1 = [Port(f"top_{i}", (xs1[i], 0), 0.5, a1) for i in range(N)]
    ports2 = [Port(f"bottom_{i}", (xs2[i], dy), 0.5, a2) for i in range(N)]

    c = gf.Component("test_facing_ports")
    routes = get_bundle(ports1, ports2)
    lengths = {}
    for i, route in enumerate(routes):
        c.add(route.references)
        lengths[i] = route.length

    if check:
        data_regression.check(lengths)
        difftest(c)

    return c


if __name__ == "__main__":

    # c = test_get_bundle(None, check=False)
    # c = test_connect_corner(None, config="A", check=False)
    # c = test_connect_corner(None, config="C", check=False) # FIXME
    # c = test_get_bundle_udirect(None, check=False)
    c = test_get_bundle_u_indirect(None, check=False, angle=90)
    # c = test_get_bundle_u_indirect(None, angle=0, check=False)
    # c = test_facing_ports(None, check=False)
    c.show()



"""Sagnac loop_mirror."""

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.mmi1x2 import mmi1x2
from gdsfactory.components.spiral_external_io import spiral_external_io
from gdsfactory.routing.manhattan import route_manhattan
from gdsfactory.types import ComponentFactory


@gf.cell
def loop_mirror(
    component: ComponentFactory = mmi1x2, bend90: ComponentFactory = bend_euler
) -> Component:
    """Returns Sagnac loop_mirror."""
    c = Component()
    component = gf.call_if_func(component)
    bend90 = gf.call_if_func(bend90)
    cref = c.add_ref(component)
    routes = route_manhattan(
        cref.ports["o3"],
        cref.ports["o2"],
        straight=gf.components.straight,
        bend=bend90,
    )
    c.add(routes.references)
    c.add_port(name="o1", port=cref.ports["o1"])
    c.absorb(cref)
    return c


@gf.cell
def loop_mirror_with_delay(loop_mirror=loop_mirror, spiral=spiral_external_io):
    """
    delay = 13e-12
    # delay = length/speed
    # length=delay*speed
    13e-12*3e8/4.2*1e6

    """
    c = Component()
    lm = c << gf.call_if_func(loop_mirror)
    s = c << spiral_external_io()

    lm.connect("o1", s.ports["o1"])
    return c


if __name__ == "__main__":
    # c = loop_mirror()
    # c = loop_mirror_rotated()
    c = loop_mirror_with_delay()
    c.show()



import pathlib

import matplotlib.pyplot as plt
import meep as mp
import numpy as np
import pandas as pd
import pydantic

import gdsfactory as gf
from gdsfactory.simulation.modes.find_modes import find_modes as find_modes_function
from gdsfactory.simulation.modes.get_mode_solver_coupler import get_mode_solver_coupler
from gdsfactory.simulation.modes.types import ModeSolverFactory
from gdsfactory.types import Callable, Optional, PathType


def coupling_length(
    neff1: float,
    neff2: float,
    power_ratio: float = 1.0,
    wavelength: float = 1.55,
) -> float:
    """
    Returns the coupling length (um) of the directional coupler
    to achieve power_ratio

    Args:
        wavelength: in um
        neff1: even supermode of the directional coupler.
        neff2: odd supermode of the directional coupler.
        power_ratio: p2/p1, where 1 means 100% power transfer

    """
    dneff = (neff1 - neff2).real
    return wavelength / (np.pi * dneff) * np.arcsin(np.sqrt(power_ratio))


@pydantic.validate_arguments
def find_coupling(
    gap: float = 0.2,
    mode_solver: ModeSolverFactory = get_mode_solver_coupler,
    find_modes: Callable = find_modes_function,
    power_ratio: float = 1.0,
    wavelength: float = 1.55,
    **kwargs
) -> float:
    """
    Returns the coupling length (um) of the directional coupler
    to achieve power_ratio

    Args:
        gap: in um
        mode_solver: function to get the mode solver
        find_modes: function to find the modes
        power_ratio: p2/p1, where 1 means 100% power transfer
        wavelength: in um

    keyword Args:
        nmodes: number of modes
        parity: for symmetries
    """
    modes = find_modes(
        mode_solver=gf.partial(mode_solver, gaps=(gap,)),
        wavelength=wavelength,
        **kwargs
    )
    ne = modes[1].neff
    no = modes[2].neff

    return coupling_length(
        neff1=ne, neff2=no, power_ratio=power_ratio, wavelength=wavelength
    )


@pydantic.validate_arguments
def find_coupling_vs_gap(
    gap1: float = 0.2,
    gap2: float = 0.4,
    steps: int = 12,
    mode_solver: ModeSolverFactory = get_mode_solver_coupler,
    find_modes: Callable = find_modes_function,
    nmodes: int = 4,
    wavelength: float = 1.55,
    parity=mp.NO_PARITY,
    filepath: Optional[PathType] = None,
    overwrite: bool = False,
    **kwargs
) -> pd.DataFrame:
    """
    Returns coupling vs gap

    Args:
        gap1:
        gap2:
        steps:
        mode_solver:
        find_modes:
        nmodes:
        wavelength:
        parity:
        filepath:
        overwrite:
    """
    if filepath and not overwrite and pathlib.Path(filepath).exists():
        return pd.read_csv(filepath)

    gaps = np.linspace(gap1, gap2, steps)
    ne = []
    no = []
    gap_to_modes = {}

    for gap in gaps:
        modes = find_modes(mode_solver=gf.partial(mode_solver, gaps=(gap,)))
        ne.append(modes[1].neff)
        no.append(modes[2].neff)
        gap_to_modes[gap] = modes

    lc = [
        coupling_length(neff1=neff1, neff2=neff2)
        for gap, neff1, neff2 in zip(gaps, ne, no)
    ]

    df = pd.DataFrame(dict(gap=gaps, ne=ne, no=no, lc=lc))
    if filepath:
        filepath = pathlib.Path(filepath)
        dirpath = filepath.parent
        dirpath.mkdir(exist_ok=True, parents=True)
        df.to_csv(filepath, index=False)
    return df


def plot_coupling_vs_gap(df: pd.DataFrame, **kwargs):
    plt.plot(df.gap, df.lc, ".-")
    plt.ylabel("100% coupling length (um)")
    plt.xlabel("gap (um)")


if __name__ == "__main__":
    df = find_coupling_vs_gap(steps=3, filepath="coupling_vs_gap.csv", overwrite=True)
    plot_coupling_vs_gap(df)
    plt.show()



from typing import Dict, Optional, Union

import meep as mp
import meep.materials as mat
import numpy as np

# Dictionary for materials with different names in Meep
MATERIAL_NAME_TO_MEEP = {
    "si": "Si",
    "sin": "Si3N4_NIR",
}

MATERIALS = [m for m in dir(mat) if not m.startswith("_")]


def get_material(
    name: str = "si",
    wavelength: float = 1.55,
    dispersive: bool = False,
    material_name_to_meep: Optional[Dict[str, Union[str, float]]] = None,
) -> mp.Medium:
    """Returns Meep Medium from database

    Args:
        name: material name
        wavelength: wavelength (um)
        dispersive: True for built-in Meep index model,
            False for simple, non-dispersive model

    Note:
        Using the built-in models can be problematic at low resolution.
        If fields are NaN or Inf, increase resolution or use a non-dispersive model
    """
    material_name_to_meep_new = material_name_to_meep or {}
    material_name_to_meep = MATERIAL_NAME_TO_MEEP.copy()
    material_name_to_meep.update(**material_name_to_meep_new)

    materials_lower = [material.lower() for material in MATERIALS]

    if name in MATERIAL_NAME_TO_MEEP.keys():
        name = MATERIAL_NAME_TO_MEEP[name]
    if name.lower() not in materials_lower:
        raise ValueError(f"material, name = {name!r} not in {MATERIALS}")

    name = MATERIALS[materials_lower.index(name.lower())]
    medium = getattr(mat, name)

    if dispersive:
        return medium
    else:
        return mp.Medium(epsilon=medium.epsilon(1 / wavelength)[0][0])


def get_index(
    wavelength: float = 1.55, name: str = "Si", dispersive: bool = False
) -> float:
    """Returns refractive index from Meep's material database.

    Args:
        name: material name
        wavelength: wavelength (um)
        dispersive: True for built-in Meep index model,
            False for simple, non-dispersive model

    Note:
        Using the built-in models can be problematic at low resolution.
        If fields are NaN or Inf, increase resolution or use a non-dispersive model

    """
    medium = get_material(name=name, wavelength=wavelength, dispersive=dispersive)

    epsilon_matrix = medium.epsilon(1 / wavelength)
    epsilon11 = epsilon_matrix[0][0]
    return epsilon11 ** 0.5


def test_index():
    n = get_index(name="sin")
    assert np.isclose(n, 1.9962797317138816)


# si = partial(get_index, name="Si")
# sio2 = partial(get_index, name="SiO2")


if __name__ == "__main__":
    print(get_index(name="sin"))



import gdsfactory as gf


@gf.cell
def straight_with_pins(**kwargs):
    c = gf.Component()
    ref = c << gf.c.straight()
    c.add_ports(ref.ports)
    return c


def test_name_with_decorator():
    c = gf.Component("test_name_with_decorator")
    c1 = c << straight_with_pins(decorator=gf.add_padding_container)
    c2 = c << straight_with_pins()

    c1.movey(-10)
    c2.movey(100)

    cells = c.get_dependencies()
    cell_names = [cell.name for cell in list(cells)]
    cell_names_unique = set(cell_names)

    if len(cell_names) != len(set(cell_names)):
        for cell_name in cell_names_unique:
            cell_names.remove(cell_name)

        cell_names_duplicated = "\n".join(set(cell_names))
        raise ValueError(
            f"Duplicated cell names in {c.name!r}:\n{cell_names_duplicated}"
        )

    referenced_cells = list(c.get_dependencies(recursive=True))
    all_cells = [c] + referenced_cells

    no_name_cells = [cell.name for cell in all_cells if cell.name.startswith("Unnamed")]
    assert (
        len(no_name_cells) == 0
    ), f"Component {c.name!r} contains {len(no_name_cells)} Unnamed cells"


if __name__ == "__main__":
    c = gf.Component("test_name_with_decorator")
    c1 = c << straight_with_pins(decorator=gf.add_padding_container)
    c2 = c << straight_with_pins()
    c1.movey(-10)
    c2.movey(100)

    cells = c.get_dependencies()
    cell_names = [cell.name for cell in list(cells)]
    cell_names_unique = set(cell_names)

    if len(cell_names) != len(set(cell_names)):
        for cell_name in cell_names_unique:
            cell_names.remove(cell_name)

        cell_names_duplicated = "\n".join(set(cell_names))
        raise ValueError(
            f"Duplicated cell names in {c.name!r}:\n{cell_names_duplicated}"
        )

    referenced_cells = list(c.get_dependencies(recursive=True))
    all_cells = [c] + referenced_cells

    no_name_cells = [cell.name for cell in all_cells if cell.name.startswith("Unnamed")]
    assert (
        len(no_name_cells) == 0
    ), f"Component {c.name!r} contains {len(no_name_cells)} Unnamed cells"



from typing import Tuple

import numpy as np
from numpy import cos, pi, sin

import gdsfactory as gf
from gdsfactory.component import Component


@gf.cell
def circle(
    radius: float = 10.0,
    angle_resolution: float = 2.5,
    layer: Tuple[int, int] = gf.LAYER.WG,
) -> Component:
    """Generate a circle geometry.

    Args:
        radius: of the circle.
        angle_resolution: number of degrees per point.
        layer: layer.

    """

    c = Component()
    t = np.linspace(0, 360, int(360 / angle_resolution) + 1) * pi / 180
    xpts = (radius * cos(t)).tolist()
    ypts = (radius * sin(t)).tolist()
    c.add_polygon(points=(xpts, ypts), layer=layer)
    return c


if __name__ == "__main__":
    c = circle()
    c.show()



from typing import Optional, Tuple

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.tech import LAYER


@gf.cell
def via(
    size: Tuple[float, float] = (0.7, 0.7),
    spacing: Tuple[float, float] = (2.0, 2.0),
    enclosure: float = 1.0,
    layer: Tuple[int, int] = LAYER.VIAC,
    layers_cladding: Optional[Tuple[Tuple[int, int], ...]] = None,
    cladding_offset: float = 0,
) -> Component:
    """Rectangular via. Defaults to a square via.

    Args:
        size: in x, y direction
        spacing: pitch_x, pitch_y
        enclosure: inclusion of via
        layer: via layer
        layers_cladding:
        cladding_offset

    .. code::

        enclosure
        _________________________________________
        |<--->                                  |
        |                      size[0]          |
        |                      <----->          |
        |      ______          ______           |
        |     |      |        |      |          |
        |     |      |        |      |  size[1] |
        |     |______|        |______|          |
        |      <------------->                  |
        |         spacing[0]                    |
        |_______________________________________|

    """
    c = Component()
    c.info["spacing"] = spacing
    c.info["enclosure"] = enclosure
    c.info["size"] = size

    width, height = size
    a = width / 2
    b = height / 2
    c.add_polygon([(-a, -b), (a, -b), (a, b), (-a, b)], layer=layer)

    layers_cladding = layers_cladding or []
    a = (width + cladding_offset) / 2
    b = (height + cladding_offset) / 2
    for layer in layers_cladding:
        c.add_polygon([(-a, -b), (a, -b), (a, b), (-a, b)], layer=layer)

    return c


viac = gf.partial(via, layer=LAYER.VIAC)
via1 = gf.partial(via, layer=LAYER.VIA1, enclosure=2)
via2 = gf.partial(via, layer=LAYER.VIA2)


if __name__ == "__main__":

    c = via()
    # c.pprint()
    print(c)
    c.show()



from gdsfactory.add_grating_couplers import (
    add_grating_couplers,
    add_grating_couplers_with_loopback_fiber_array,
    add_grating_couplers_with_loopback_fiber_single,
)
from gdsfactory.add_padding import add_padding_container
from gdsfactory.add_termination import add_termination
from gdsfactory.components.bend_port import bend_port
from gdsfactory.components.cavity import cavity
from gdsfactory.components.extension import extend_ports
from gdsfactory.components.ring_single_dut import ring_single_dut
from gdsfactory.functions import rotate
from gdsfactory.routing.add_electrical_pads_shortest import add_electrical_pads_shortest
from gdsfactory.routing.add_electrical_pads_top import add_electrical_pads_top
from gdsfactory.routing.add_fiber_array import add_fiber_array
from gdsfactory.routing.add_fiber_single import add_fiber_single
from gdsfactory.routing.fanout2x2 import fanout2x2
from gdsfactory.tech import Library

container_library = Library(name="generic_containers")
container_library.register(
    [
        add_electrical_pads_shortest,
        add_electrical_pads_top,
        add_fiber_array,
        add_fiber_single,
        add_grating_couplers,
        add_grating_couplers_with_loopback_fiber_single,
        add_grating_couplers_with_loopback_fiber_array,
        add_padding_container,
        add_termination,
        cavity,
        extend_ports,
        fanout2x2,
        ring_single_dut,
        bend_port,
        rotate,
    ]
)


__all__ = list(container_library.factory.keys())



"""Compute and write Sparameters using Meep in an MPI pool."""

import multiprocessing
import pathlib
import shutil
import time
from pathlib import Path
from typing import Dict, List

import numpy as np
import pydantic

import gdsfactory as gf
from gdsfactory.config import CONFIG, logger
from gdsfactory.simulation.get_sparameters_path import (
    get_sparameters_path_meep as get_sparameters_path,
)
from gdsfactory.simulation.gmeep.write_sparameters_meep import remove_simulation_kwargs
from gdsfactory.simulation.gmeep.write_sparameters_meep_mpi import (
    write_sparameters_meep_mpi,
)
from gdsfactory.tech import LAYER_STACK, LayerStack

ncores = multiprocessing.cpu_count()


@pydantic.validate_arguments
def write_sparameters_meep_mpi_pool(
    jobs: List[Dict],
    cores_per_run: int = 2,
    total_cores: int = 4,
    temp_dir: Path = CONFIG["sparameters"] / "temp",
    delete_temp_files: bool = True,
    dirpath: Path = CONFIG["sparameters"],
    layer_stack: LayerStack = LAYER_STACK,
    **kwargs,
) -> List[Path]:
    """Write Sparameters and returns the filepaths
    Given a list of write_sparameters_meep keyword arguments (the "jobs"),
        launches them in different cores
    Each simulation is assigned "cores_per_run" cores
    A total of "total_cores" is assumed, if cores_per_run * len(jobs) > total_cores
    then the overflow will run sequentially (not in parallel)

    Args
        jobs: list of Dicts containing the simulation settings for each job.
            for write_sparameters_meep
        cores_per_run: number of processors to assign to each component simulation
        total_cores: total number of cores to use
        temp_dir: temporary directory to hold simulation files
        delete_temp_files: deletes temp_dir when done
        dirpath: directory to store Sparameters
        layer_stack:

    keyword Args:
        overwrite: overwrites stored simulation results.
        dispersive: use dispersive models for materials (requires higher resolution)
        extend_ports_length: to extend ports beyond the PML
        t_clad_top: thickness for cladding above core
        t_clad_bot: thickness for cladding below core
        tpml: PML thickness (um)
        clad_material: material for cladding
        is_3d: if True runs in 3D
        wl_min: wavelength min (um)
        wl_max: wavelength max (um)
        wl_steps: wavelength steps
        dfcen: delta frequency
        port_source_name: input port name
        port_field_monitor_name:
        port_margin: margin on each side of the port
        distance_source_to_monitors: in (um) source goes before
        port_source_offset: offset between source GDS port and source MEEP port
        port_monitor_offset: offset between monitor GDS port and monitor MEEP port

    Returns:
        filepath list for sparameters CSV (wavelengths, s11a, s12m, ...)
            where `a` is the angle in radians and `m` the module

    """
    # Parse jobs
    jobs_to_run = []
    for job in jobs:
        settings = remove_simulation_kwargs(kwargs)
        filepath = job.get(
            "filepath",
            get_sparameters_path(
                component=job["component"],
                dirpath=dirpath,
                layer_stack=layer_stack,
                **settings,
            ),
        )
        if filepath.exists():
            job.update(**kwargs)
            if job.get("overwrite", False):
                pathlib.Path.unlink(filepath)
                logger.info(
                    f"Simulation {filepath!r} found and overwrite is True. "
                    "Deleting file and adding it to the queue."
                )
                jobs_to_run.append(job)
            else:
                logger.info(
                    f"Simulation {filepath!r} found exists and "
                    "overwrite is False. Removing it from the queue."
                )
        else:
            logger.info(f"Simulation {filepath!r} not found. Adding it to the queue")
            jobs_to_run.append(job)

    # Update jobs
    jobs = jobs_to_run

    # Setup pools
    num_pools = int(np.ceil(cores_per_run * len(jobs) / total_cores))
    jobs_per_pool = int(np.floor(total_cores / cores_per_run))
    njobs = len(jobs)

    logger.info(f"Running parallel simulations over {njobs} jobs")
    logger.info(
        f"Using a total of {total_cores} cores with {cores_per_run} cores per job"
    )
    logger.info(
        f"Tasks split amongst {num_pools} pools with up to {jobs_per_pool} jobs each."
    )

    i = 0
    # For each pool
    for j in range(num_pools):
        filepaths = []

        # For each job in the pool
        for k in range(jobs_per_pool):
            # Flag to catch non full pools
            if i >= njobs:
                continue
            logger.info(f"Task {k} of pool {j} is job {i}")

            # Obtain current job
            simulations_settings = jobs[i]

            filepath = write_sparameters_meep_mpi(
                cores=cores_per_run,
                temp_dir=temp_dir,
                temp_file_str=f"write_sparameters_meep_mpi_{i}",
                wait_to_finish=False,
                **simulations_settings,
            )
            filepaths.append(filepath)

            # Increment task number
            i += 1

        # Wait for pool to end
        done = False
        num_pool_jobs = len(filepaths)
        while not done:
            # Check if all jobs finished
            jobs_done = 0
            for filepath in filepaths:
                if filepath.exists():
                    jobs_done += 1
            if jobs_done == num_pool_jobs:
                done = True
            else:
                time.sleep(1)

    if delete_temp_files:
        shutil.rmtree(temp_dir)
    return filepaths


if __name__ == "__main__":

    # Multicore pools example
    c1 = gf.c.straight(length=5)
    p = 3
    c1 = gf.add_padding_container(c1, default=0, top=p, bottom=p)

    c2 = gf.c.straight(length=4)
    p = 3
    c2 = gf.add_padding_container(c2, default=0, top=p, bottom=p)

    c1_dict = {
        "component": c1,
        "run": True,
        "overwrite": True,
        "lazy_parallelism": True,
        "filepath": Path("c1_dict.csv"),
    }
    c2_dict = {
        "component": c2,
        "run": True,
        "overwrite": False,
        "lazy_parallelism": True,
        "filepath": Path("c2_dict.csv"),
    }
    c3_dict = {
        "component": c2,
        "run": True,
        "overwrite": True,
        "lazy_parallelism": True,
        "resolution": 40,
        "port_source_offset": 0.3,
        "filepath": Path("c3_dict.csv"),
    }

    # jobs
    jobs = [
        c1_dict,
        c2_dict,
        c3_dict,
    ]

    filepaths = write_sparameters_meep_mpi_pool(
        jobs=jobs,
        cores_per_run=4,
        total_cores=10,
        delete_temp_files=False,
    )



import pytest
from pytest_regressions.data_regression import DataRegressionFixture

from gdsfactory.components.mmi2x2 import mmi2x2
from gdsfactory.components.mzi_phase_shifter import mzi_phase_shifter_top_heater_metal

# from gdsfactory.components.spiral_inner_io import spiral_inner_io
from gdsfactory.containers import container_library
from gdsfactory.difftest import difftest

container_factory = container_library.factory
container_names = set(container_factory.keys()) - set(
    ["add_grating_couplers_wi_e642a51b"]
)
component = mzi_phase_shifter_top_heater_metal(splitter=mmi2x2)


@pytest.mark.parametrize("container_type", container_names)
def test_settings(container_type: str, data_regression: DataRegressionFixture) -> None:
    """Avoid regressions when exporting settings."""
    c = container_factory[container_type](component=component)
    data_regression.check(c.to_dict())


@pytest.mark.parametrize("container_type", container_names)
def test_gds(container_type: str) -> None:
    """Avoid regressions in GDS geometry shapes and layers."""
    c = container_factory[container_type](component=component)
    difftest(c)


# Special test cases for exotic components
# def test_add_gratings_and_loopback(data_regression: DataRegressionFixture) -> None:
#     """This container requires all ports to face the same direction."""
#     c = add_gratings_and_loopback(component=spiral_inner_io())
#     data_regression.check(c.settings)


if __name__ == "__main__":
    for i in container_names:
        print(i)



from typing import Optional

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.coupler90 import coupler90 as coupler90function
from gdsfactory.components.coupler_straight import (
    coupler_straight as coupler_straight_function,
)
from gdsfactory.cross_section import strip
from gdsfactory.snap import assert_on_2nm_grid
from gdsfactory.types import ComponentFactory, CrossSectionFactory


@gf.cell
def coupler_ring(
    gap: float = 0.2,
    radius: float = 5.0,
    length_x: float = 4.0,
    coupler90: ComponentFactory = coupler90function,
    bend: Optional[ComponentFactory] = None,
    coupler_straight: ComponentFactory = coupler_straight_function,
    cross_section: CrossSectionFactory = strip,
    bend_cross_section: Optional[CrossSectionFactory] = None,
    **kwargs
) -> Component:
    r"""Coupler for ring.

    Args:
        gap: spacing between parallel coupled straight waveguides.
        radius: of the bends.
        length_x: length of the parallel coupled straight waveguides.
        coupler90: straight coupled to a 90deg bend.
        bend: factory for bend
        coupler_straight: two parallel coupled straight waveguides.
        cross_section:
        kwargs: cross_section settings

    .. code::

           2             3
           |             |
            \           /
             \         /
           ---=========---
         1    length_x    4


    """
    bend = bend or bend_euler

    c = Component()
    assert_on_2nm_grid(gap)

    # define subcells
    coupler90_component = (
        coupler90(
            gap=gap,
            radius=radius,
            bend=bend,
            cross_section=cross_section,
            bend_cross_section=bend_cross_section,
            **kwargs
        )
        if callable(coupler90)
        else coupler90
    )
    coupler_straight_component = (
        coupler_straight(
            gap=gap, length=length_x, cross_section=cross_section, **kwargs
        )
        if callable(coupler_straight)
        else coupler_straight
    )

    # add references to subcells
    cbl = c << coupler90_component
    cbr = c << coupler90_component
    cs = c << coupler_straight_component

    # connect references
    y = coupler90_component.y
    cs.connect(port="o4", destination=cbr.ports["o1"])
    cbl.reflect(p1=(0, y), p2=(1, y))
    cbl.connect(port="o2", destination=cs.ports["o2"])

    c.absorb(cbl)
    c.absorb(cbr)
    c.absorb(cs)

    c.add_port("o1", port=cbl.ports["o3"])
    c.add_port("o2", port=cbl.ports["o4"])
    c.add_port("o3", port=cbr.ports["o3"])
    c.add_port("o4", port=cbr.ports["o4"])
    c.auto_rename_ports()
    return c


if __name__ == "__main__":

    c = coupler_ring(width=1, layer=(2, 0))
    c.show(show_subports=True)



"""
In this example we change the layer of the straight
"""

import gdsfactory as gf


@gf.cell
def wg(layer=(2, 0), **kwargs):
    return gf.components.straight(layer=layer, **kwargs)


if __name__ == "__main__":
    c = wg()
    c.show()



import itertools
import math
from typing import Tuple

import klayout.db as pya
import pyqtree
from klayout.dbcore import Cell

import gdsfactory.autoplacer.functions as ap
from gdsfactory.autoplacer.cell_list import CellList
from gdsfactory.autoplacer.library import Library


class AutoPlacer(pya.Layout):
    """klayout autoplacer

    Args:
        name: name of the container
        max_width: 1e8 (nm)
        max_height: 1e8 (nm)

    automatically packs and position cells. You feed it cells that are loaded and padded using `library`.

    Instantiate an AutoPlacer with a `name`, `max_width` and `max_height` properties. If this Autoplacer is the main top-level cell of the mask, `max_width` and `max_height` should be the desired width and height of the mask.

    You can then pack cells into the AutoPlacer using:

    - `AutoPlacer.pack_auto(cell, origin, direction)` automatically finds a location for the component.
        You can provide an `origin` argument, which can take any value from `autoplacer.SOUTH_WEST`, `autoplacer.SOUTH_EAST`, `autoplacer.NORTH_WEST`, `autoplacer.NORTH_EAST`.
        The autoplacer will search for a place to put the cell starting from that cardinal point. It can search in either `VERTICAL` or `HORIZONTAL` direction.
    - `AutoPlacer.pack_manual(cell, x, y, origin)` allows you to manually place a component at an absolute location.
    - `AutoPlacer.pack_many` behaves like `AutoPlacer.pack_auto` but accepts a list of devices.

    """

    def __init__(
        self, name: str, max_width: float = 1e8, max_height: float = 1e8
    ) -> None:
        """constructor"""
        # Construct
        super(AutoPlacer, self).__init__()

        # get width and height
        self.max_width = max_width
        self.max_height = max_height
        global COUNTER, AUTOPLACER_REGISTRY
        if name in ap.AUTOPLACER_REGISTRY:
            self.name = "{}_{}".format(name, ap.COUNTER)
            ap.COUNTER += 1
        else:
            self.name = name
        ap.AUTOPLACER_REGISTRY[self.name] = self

        # Register
        ap.WORKING_MEMORY["__AutoPlacer_{}".format(self.name)] = self

        # Create the quadtree which will enable efficient queries
        bbox = (0, 0, self.max_width, self.max_height)
        self.quadtree = pyqtree.Index(bbox=bbox)

        # Make a topcell
        self.create_cell(self.name)

    def get_edge(self, collisions, bbox, direction):
        """Get a particular edge of some collisions"""
        if direction == ap.NORTH:
            return max(n for (w, s, e, n) in collisions) + ap.GRID
        elif direction == ap.SOUTH:
            return min(s for (w, s, e, n) in collisions) - ap.GRID - bbox.height()
        elif direction == ap.EAST:
            return min(e for (w, s, e, n) in collisions) + ap.GRID
        elif direction == ap.WEST:
            return max(w for (w, s, e, n) in collisions) - ap.GRID - bbox.width()

    def import_cell(self, cell: Cell) -> Cell:
        """Imports a cell from another Layout"""
        # If the cell is already in the library, skip loading
        if self.cell(cell.name):
            return self.cell(cell.name)

        # Create a holder cell and copy in the shapes
        new_cell = self.create_cell(cell.name)
        new_cell.copy_shapes(cell)

        # Import all the child cells
        for child_index in cell.each_child_cell():
            self.import_cell(cell.layout().cell(child_index))

        # Import all of the instances, doing the mapping from Layout to Layout
        for instance in cell.each_inst():
            cell_index = self.cell(instance.cell.name).cell_index()
            # new_instance = pya.CellInstArray(cell_index, instance.trans)
            new_instance = instance.cell_inst.dup()
            new_instance.cell_index = cell_index
            new_cell.insert(new_instance)
        return new_cell

    def inside(self, bbox):
        """Check that something is inside the mask"""
        mask = pya.Box(0, 0, self.max_width + 1, self.max_height + 1)
        return mask.contains(bbox)

    def find_collisions(self, bbox, sx, sy):
        """Test for intersection of objects"""
        box = sx, sy, sx + bbox.width(), sy + bbox.height()
        return self.quadtree.intersect(box)

    def compute_start(self, cell, origin):
        """Figure out where the search should start for a given cell"""
        w = cell.bbox().width()
        h = cell.bbox().height()
        if origin == ap.SOUTH_WEST:
            return [0, 0]
        elif origin == ap.SOUTH_EAST:
            return [self.max_width - w, 0]
        elif origin == ap.NORTH_WEST:
            return [0, self.max_height - h]
        elif origin == ap.NORTH_EAST:
            return [self.max_width - w, self.max_height - h]

    def find_space(self, cell, origin=ap.SOUTH_WEST, direction=ap.VERTICAL):
        """Find space for a cell by brute-force search"""
        if direction == ap.VERTICAL:
            return self.find_space_vertical(cell, origin)
        elif direction == ap.HORIZONTAL:
            return self.find_space_horizontal(cell, origin)

    def find_space_vertical(self, cell, origin=ap.SOUTH_WEST):
        """Find space for a cell by brute-force search - horizontal"""
        # Compute boundaries
        bbox = cell.bbox()
        x_bound = self.max_width - bbox.width()
        y_bound = self.max_height - bbox.height()

        # Compute start position
        start = self.compute_start(cell, origin)
        sx = start[0]
        while sx >= 0 and sx <= x_bound:
            sy = start[1]
            edge = []
            while sy >= 0 and sy <= y_bound:
                # Find any collisions with cell that have been packed
                collisions = self.find_collisions(bbox, sx, sy)
                if len(collisions) == 0:
                    return sx, sy

                # Otherwise, search from the north most collision
                sy = self.get_edge(
                    collisions, bbox, ap.NORTH if ap.SOUTH in origin else ap.SOUTH
                )
                e = self.get_edge(
                    collisions, bbox, ap.EAST if ap.WEST in origin else ap.WEST
                )
                edge.append(e)

            if not edge:
                return
            sx = min(edge) if ap.WEST in origin else max(edge)

    def find_space_horizontal(self, cell, origin=ap.SOUTH_WEST):
        """Find space for a cell by brute-force search - vertical"""
        # Compute boundaries
        bbox = cell.bbox()
        x_bound = self.max_width - bbox.width()
        y_bound = self.max_height - bbox.height()

        # Compute start position
        start = self.compute_start(cell, origin)
        sy = start[1]
        while sy >= 0 and sy <= y_bound:
            sx = start[0]
            edge = []
            while sx >= 0 and sx <= x_bound:
                # Find any collisions with cell that have been packed
                collisions = self.find_collisions(bbox, sx, sy)
                if len(collisions) == 0:
                    return sx, sy

                # Otherwise, search from the northmost collision
                sx = self.get_edge(
                    collisions, bbox, ap.EAST if ap.WEST in origin else ap.WEST
                )
                e = self.get_edge(
                    collisions, bbox, ap.NORTH if ap.SOUTH in origin else ap.SOUTH
                )
                edge.append(e)

            if not edge:
                return
            sy = min(edge) if ap.SOUTH in origin else max(edge)

    def pack_manual(
        self,
        cell: Cell,
        x: float,
        y: float,
        origin: Tuple[int, int] = ap.SOUTH_WEST,
        tboxes: None = None,
    ) -> None:
        """
        Pack a cell at a manually selected position
        """
        box = cell.bbox()

        # Do an offset if origin is given
        w = box.width()
        h = box.height()
        oy, ox = origin
        x -= {ap.WEST: 0, ap.EAST: w, ap.MIDDLE: w / 2}[ox]
        y -= {ap.SOUTH: 0, ap.NORTH: w, ap.MIDDLE: h / 2}[oy]

        # Insert into the quadtree

        if tboxes is None:
            tbox = (x, y, x + box.width(), y + box.height())
            tboxes = [tbox]

        for tbox in tboxes:
            self.quadtree.insert(tbox, tbox)

        new_cell = self.import_cell(cell)

        # Make an instance
        transform = pya.Trans(int(x - box.left), int(y - box.bottom))
        new_instance = pya.CellInstArray(new_cell.cell_index(), transform)
        self.cell(self.name).insert(new_instance)

    def pack_auto(self, cell, origin=ap.SOUTH_WEST, direction=ap.VERTICAL):
        """
        Pack a cell automatically
        """
        coordinate = self.find_space(cell, origin, direction)
        if coordinate is None:
            raise ap.OutOfSpaceError(
                "Out of space for {} ({} x {} mm) in {}".format(
                    cell.name,
                    cell.bbox().width() / 1e6,
                    cell.bbox().height() / 1e6,
                    self.name,
                )
            )
        else:
            self.pack_manual(cell, *coordinate)

    def pack_many(self, cells, origin=ap.SOUTH_WEST, direction=ap.VERTICAL):
        """
        Pack many cells with the same settings
        """
        failed = []
        for cell in cells:
            try:
                self.pack_auto(cell, origin, direction)
            except ap.OutOfSpaceError as e:
                print(e)
                failed.append(cell)
        return CellList(failed)

    def shrink(self):
        """Shrink-wrap"""
        bbox = self.top_cell().bbox()
        self.max_width = bbox.width()
        self.max_height = bbox.height()

    def draw_boundary(self, layer: int = ap.DEVREC_LAYER) -> None:
        """Draw a box into the topcell"""
        layer = self.layer(layer, 0)
        self.cell(self.name).shapes(layer).insert(
            pya.Box(0, 0, self.max_width, self.max_height)
        )

    def write(self, *args, **kwargs) -> None:
        """Draw boundary on write"""
        if not kwargs.get("shrink", False):
            self.draw_boundary()

        super(AutoPlacer, self).write(*args, **kwargs)

    def pack_groups(
        self,
        cells,
        cols=None,
        rows=None,
        aspect=3,
        dimension=ap.WIDTH,
        normalization_origin=ap.SOUTH_WEST,
        direction=ap.VERTICAL,
        granularity=ap.GRANULARITY,
        name=None,
        origin=ap.SOUTH_WEST,
    ):
        """Pack column-wise"""
        for group in cells.groups(dimension, granularity):
            self.pack_grid(
                group,
                cols,
                rows,
                aspect=aspect,
                normalization_origin=normalization_origin,
                name=name,
                origin=origin,
            )

    def pack_grid(
        self,
        cells: CellList,
        cols: None = None,
        rows: None = None,
        aspect: int = 3,
        direction: int = ap.VERTICAL,
        align: int = ap.BOTH,
        normalization_origin: Tuple[int, int] = ap.SOUTH_WEST,
        origin: Tuple[int, int] = ap.SOUTH_WEST,
        name: None = None,
        padding: int = ap.PADDING,
    ) -> None:
        """
        Pack into a grid, assuming that all cells are the same size
        """
        if not cells:
            return

        # Get the name from the longest common substring
        if name is None:
            name = ap.longest_common_prefix([c.name for c in cells]) + "_g"
            name = name if len(name) > 2 else "Misc"

        # Make the block with approriate size
        block = AutoPlacer(name)

        # Collect and clean the cells
        cells.align(align)
        cells.normalize(ap.BOTH, normalization_origin)
        cells.pad(padding=padding)
        # cells.sort(lambda c: c.name)

        # Safety check
        assert not (
            cols is not None and rows is not None
        ), "Don't specify both rows and cols"

        # Infer the dimensions, targeting an aspect ratio of 1
        if rows is None and cols is None:
            b = cells[0].bbox()
            cols = ap.estimate_cols(len(cells), b.width(), b.height(), aspect)

        # Accept user parameters
        if cols is None:
            cols = int(math.ceil(len(cells) / rows))
        elif rows is None:
            rows = int(math.ceil(len(cells) / cols))

        # Get dimensions
        bbox = cells[0].bbox()
        w, h = bbox.width(), bbox.height()

        # Place
        assert cols * rows >= len(cells)
        grid = itertools.product(list(range(int(cols))), list(range(int(rows))))
        for cell, (col, row) in zip(cells, grid):
            block.pack_manual(cell, w * col, h * row)

        # Shrink and add a boundary
        block.shrink()
        block.draw_boundary()

        # Pack into the container
        self.pack_auto(block.top_cell(), direction=direction, origin=origin)

    def pack_random(
        self,
        cells,
        origin=ap.SOUTH_WEST,
        direction=ap.HORIZONTAL,
        normalization_origin=ap.SOUTH_WEST,
        add_padding=True,
        padding=ap.PADDING,
    ):
        """Pack at random"""
        if add_padding:
            cells.pad(padding=padding)
        return self.pack_many(cells, direction=direction, origin=origin)

    def pack_lumped(self, cells, width, height, origin=ap.SOUTH_WEST):
        """Pack at random in a lump"""
        # Get the name from the longest common substring
        name = ap.longest_common_prefix([c.name for c in cells]) + "_l"
        name = name if len(name) > 2 else "Misc"

        # Make an autoplacer
        lump = AutoPlacer(name, width, height)
        lump.pack_random(cells)
        lump.shrink()
        lump.draw_boundary()
        return self.pack_auto(lump.top_cell(), origin=origin)

    def pack_corners(self, component):
        for corner in ap.CORNERS:
            self.pack_auto(component, corner)

    def estimate_size(self, cells):
        """Estimate the size of the placer"""
        # Compute squarish dimensions
        N = len(cells)
        rows = math.ceil(math.sqrt(N))
        cols = math.floor(math.sqrt(N))

        # Analyze BBoxes
        bboxes = [cell.bbox() for cell in cells]
        max_width = max(b.width() for b in bboxes)
        max_height = max(b.height() for b in bboxes)

        # Set new max height and width
        return max_width * cols + ap.PADDING, max_height * rows + ap.PADDING


if __name__ == "__main__":
    lib = Library()
    mask = AutoPlacer("mask", 25e6, 25e6)
    mask.pack_grid(lib.pop("align"))
    mask.pack_grid(lib.pop(".*"))
    mask.write("build/mask.gds")



from click.testing import CliRunner

from gdsfactory import __version__
from gdsfactory.gf import gf


def test_gf_version() -> None:
    """checks that the CLI returns the correct version"""
    runner = CliRunner()
    result = runner.invoke(gf, ["--version"])

    # print(result.output)
    # print(__version__)
    assert result.exit_code == 0
    assert result.output.startswith(__version__)


if __name__ == "__main__":
    test_gf_version()



import gdsfactory as gf


def from_dphox(device: "dp.Device", foundry: "dp.foundry.Foundry") -> gf.Component:
    """Converts a Dphox Device into a gdsfactory Component.

    Note that you need to install dphox `pip install dphox`

    https://dphox.readthedocs.io/en/latest/index.html

    Args:
        device:
        foundry:
    """
    c = gf.Component(device.name)

    for layer_name, shapely_multipolygon in device.layer_to_polys.items():
        for poly in shapely_multipolygon:
            layer = foundry.layer_to_gds_label[layer_name]
            c.add_polygon(points=poly, layer=layer)

    for port_name, port in device.port.items():
        c.add_port(
            name=port_name,
            midpoint=(port.x, port.y),
            orientation=port.a,
            width=port.w,
        )
    return c


if __name__ == "__main__":
    import dphox as dp
    from dphox.demo import lateral_nems_ps

    nems_ps = lateral_nems_ps(waveguide_w=0.3)

    c = from_dphox(nems_ps, foundry=dp.foundry.FABLESS)
    c.show()



import gdsfactory as gf
from gdsfactory.add_padding import get_padding_points
from gdsfactory.component import Component
from gdsfactory.cross_section import strip
from gdsfactory.path import arc, extrude
from gdsfactory.snap import snap_to_grid
from gdsfactory.types import CrossSectionOrFactory


@gf.cell
def bend_circular(
    angle: int = 90,
    npoints: int = 720,
    with_cladding_box: bool = True,
    cross_section: CrossSectionOrFactory = strip,
    **kwargs
) -> Component:
    """Returns a radial arc.

    Args:
        angle: angle of arc (degrees)
        npoints: number of points
        with_cladding_box: square in layers_cladding to remove DRC
        cross_section:
        kwargs: cross_section settings


    .. code::

                  o2
                  |
                 /
                /
               /
       o1_____/

    """
    x = cross_section(**kwargs) if callable(cross_section) else cross_section
    radius = x.info["radius"]

    p = arc(radius=radius, angle=angle, npoints=npoints)
    c = Component()
    path = extrude(p, x)
    ref = c << path
    c.add_ports(ref.ports)

    c.info.length = snap_to_grid(p.length())
    c.info.dy = float(abs(p.points[0][0] - p.points[-1][0]))
    c.info.radius = float(radius)

    if with_cladding_box and x.info["layers_cladding"]:
        layers_cladding = x.info["layers_cladding"]
        cladding_offset = x.info["cladding_offset"]
        top = cladding_offset if angle == 180 else 0
        points = get_padding_points(
            component=c,
            default=0,
            bottom=cladding_offset,
            right=cladding_offset,
            top=top,
        )
        for layer in layers_cladding or []:
            c.add_polygon(points, layer=layer)

    c.absorb(ref)
    return c


bend_circular180 = gf.partial(bend_circular, angle=180)


if __name__ == "__main__":
    c = bend_circular(width=2, layer=gf.LAYER.M1)
    # c = bend_circular(cross_section=gf.cross_section.pin, radius=5)
    c.pprint()
    c.show()

    c = bend_circular180()
    c.plot("qt")

    # from phidl.quickplotter import quickplot2
    # c = bend_circular_trenches()
    # c = bend_circular_deep_rib()
    # print(c.ports)
    # print(c.length, np.pi * 10)
    # print(c.ports.keys())
    # print(c.ports['o2'].midpoint)
    # print(c.settings)
    # c = bend_circular_slot()
    # c = bend_circular(width=0.45, radius=5)
    # c.plot()
    # quickplot2(c)



import numpy as np

from gdsfactory.simulation.modes.find_modes import find_modes
from gdsfactory.simulation.modes.get_mode_solver_rib import get_mode_solver_rib


def test_find_modes():
    ms = get_mode_solver_rib(wg_width=0.45)
    modes = find_modes(mode_solver=ms)
    m1 = modes[1]
    m2 = modes[2]
    # print(m1.neff)
    # print(m2.neff)

    neff1 = 2.329468837266023
    neff2 = 1.7128271512313074

    assert np.isclose(m1.neff, neff1), (m1.neff, neff1)
    assert np.isclose(m2.neff, neff2), (m2.neff, neff2)


if __name__ == "__main__":
    test_find_modes()
    # ms = get_mode_solver_rib(wg_width=0.45)
    # modes = find_neff(mode_solver=ms)
    # m1 = modes[1]
    # m2 = modes[2]
    # print(m1.neff)
    # print(m2.neff)
    # neff1 = 2.342628111145838
    # neff2 = 1.7286034634949181

    # assert np.isclose(m1.neff, neff1), (m1.neff, neff1)
    # assert np.isclose(m2.neff, neff2), (m2.neff, neff2)



"""Compute and write Sparameters using Meep in MPI."""

import multiprocessing
import pathlib
import pickle
import shlex
import subprocess
import time
from pathlib import Path
from typing import Optional

import pydantic

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.config import CONFIG, logger
from gdsfactory.simulation.get_sparameters_path import (
    get_sparameters_path_meep as get_sparameters_path,
)
from gdsfactory.simulation.gmeep.write_sparameters_meep import remove_simulation_kwargs
from gdsfactory.tech import LAYER_STACK, LayerStack

ncores = multiprocessing.cpu_count()


@pydantic.validate_arguments
def write_sparameters_meep_mpi(
    component: Component,
    cores: int = ncores,
    filepath: Optional[Path] = None,
    dirpath: Path = CONFIG["sparameters"],
    layer_stack: LayerStack = LAYER_STACK,
    temp_dir: Path = CONFIG["sparameters"] / "temp",
    temp_file_str: str = "write_sparameters_meep_mpi",
    overwrite: bool = False,
    wait_to_finish: bool = True,
    **kwargs,
) -> Path:
    """Write Sparameters using multiple cores and MPI
    and returns Sparameters CSV filepath.

    Simulates each time using a different input port (by default, all of them)
    unless you specify port_symmetries:

    port_symmetries = {"o1":
            {
                "s11": ["s22","s33","s44"],
                "s21": ["s21","s34","s43"],
                "s31": ["s13","s24","s42"],
                "s41": ["s14","s23","s32"],
            }
        }

    Args:
        component: gdsfactory Component.
        cores: number of processors.
        filepath: to store pandas Dataframe with Sparameters in CSV format.
            Defaults to dirpath/component_.csv
        dirpath: directory to store Sparameters
        layer_stack:
        temp_dir: temporary directory to hold simulation files.
        temp_file_str: names of temporary files in temp_dir.
        overwrite: overwrites stored simulation results.
        wait_to_finish:

    Keyword Args:
        resolution: in pixels/um (20: for coarse, 120: for fine)
        port_symmetries: Dict to specify port symmetries, to save number of simulations
        source_ports: list of port string names to use as sources
        dirpath: directory to store Sparameters
        layer_stack: LayerStack class
        port_margin: margin on each side of the port
        port_monitor_offset: offset between monitor GDS port and monitor MEEP port
        port_source_offset: offset between source GDS port and source MEEP port
        filepath: to store pandas Dataframe with Sparameters in CSV format.
        animate: saves a MP4 images of the simulation for inspection, and also
            outputs during computation. The name of the file is the source index
        lazy_parallelism: toggles the flag "meep.divide_parallel_processes" to
            perform the simulations with different sources in parallel
        dispersive: use dispersive models for materials (requires higher resolution)
        extend_ports_length: to extend ports beyond the PML
        layer_stack: Dict of layer number (int, int) to thickness (um)
        t_clad_top: thickness for cladding above core
        t_clad_bot: thickness for cladding below core
        tpml: PML thickness (um)
        clad_material: material for cladding
        is_3d: if True runs in 3D
        wl_min: wavelength min (um)
        wl_max: wavelength max (um)
        wl_steps: wavelength steps
        dfcen: delta frequency
        port_source_name: input port name
        port_field_monitor_name:
        port_margin: margin on each side of the port
        distance_source_to_monitors: in (um) source goes before
        port_source_offset: offset between source GDS port and source MEEP port
        port_monitor_offset: offset between monitor GDS port and monitor MEEP port

    Returns:
        filepath for sparameters CSV (wavelengths, s11a, s12m, ...)
            where `a` is the angle in radians and `m` the module
    """
    settings = remove_simulation_kwargs(kwargs)
    filepath = filepath or get_sparameters_path(
        component=component,
        dirpath=dirpath,
        layer_stack=layer_stack,
        **settings,
    )
    filepath = pathlib.Path(filepath)
    if filepath.exists() and not overwrite:
        logger.info(f"Simulation {filepath!r} already exists")
        return filepath

    # Save the component object to simulation for later retrieval
    temp_dir.mkdir(exist_ok=True, parents=True)
    tempfile = temp_dir / temp_file_str
    component_file = tempfile.with_suffix(".pkl")
    kwargs.update(filepath=str(filepath))

    with open(component_file, "wb") as outp:
        pickle.dump(component, outp, pickle.HIGHEST_PROTOCOL)

    # Write execution file
    script_lines = [
        "import pickle\n",
        "from gdsfactory.simulation.gmeep import write_sparameters_meep\n\n",
        'if __name__ == "__main__":\n\n',
        f"\twith open(\"{component_file}\", 'rb') as inp:\n",
        "\t\tcomponent = pickle.load(inp)\n\n"
        "\twrite_sparameters_meep(component = component,\n",
    ]
    for key in kwargs.keys():
        script_lines.append(f"\t\t{key} = {kwargs[key]!r},\n")

    script_lines.append("\t)")
    script_file = tempfile.with_suffix(".py")
    script_file_obj = open(script_file, "w")
    script_file_obj.writelines(script_lines)
    script_file_obj.close()

    command = f"mpirun -np {cores} python {script_file}"
    logger.info(command)
    logger.info(str(filepath))

    subprocess.Popen(
        shlex.split(command),
        shell=False,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    if wait_to_finish:
        while not filepath.exists():
            time.sleep(1)

    return filepath


if __name__ == "__main__":
    c1 = gf.c.straight(length=5)
    p = 3
    c1 = gf.add_padding_container(c1, default=0, top=p, bottom=p)

    instance_dict = {
        "component": c1,
        "run": True,
        "overwrite": True,
        "lazy_parallelism": True,
        "filepath": "instance_dict.csv",
    }

    proc = write_sparameters_meep_mpi(
        instance=instance_dict,
        cores=3,
        verbosity=True,
    )



from typing import Callable, Optional

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.taper_cross_section import taper_cross_section
from gdsfactory.cross_section import strip
from gdsfactory.port import select_ports_optical
from gdsfactory.types import ComponentFactory, CrossSectionFactory


@cell
def add_tapers(
    component: Component,
    taper: ComponentFactory = taper_cross_section,
    select_ports: Optional[Callable] = select_ports_optical,
    taper_port_name1: str = "o1",
    taper_port_name2: str = "o2",
    cross_section2: CrossSectionFactory = strip,
    **kwargs
) -> Component:
    """returns new component with taper in all optical ports.

    Args:
        component:
        taper: factory for taper
        select_ports:
        taper_port_name1:
        taper_port_name2:
        cross_section2: end cross_section factory (cross_section)

    Keyword Args:
        cross_section1: start cross_section factory
        length: transition length
        npoints: number of points
        linear: shape of the transition, sine when False
        kwargs: cross_section settings for section2
    """

    c = gf.Component()
    ports_to_taper = select_ports(component.ports) if select_ports else component.ports
    ports_to_taper_names = [p.name for p in ports_to_taper.values()]

    for port_name, port in component.ports.items():
        if port.name in ports_to_taper_names:
            taper_ref = c << taper(
                cross_section1=port.cross_section,
                cross_section2=cross_section2,
                **kwargs
            )
            taper_ref.connect(taper_ref.ports[taper_port_name1].name, port)
            c.add_port(name=port_name, port=taper_ref.ports[taper_port_name2])
        else:
            c.add_port(name=port_name, port=port)
    c.add_ref(component)
    c.copy_child_info(component)
    return c


if __name__ == "__main__":
    c0 = gf.components.straight(width=2, cross_section=gf.cross_section.rib)
    xs_rib_tip = gf.cross_section.strip_rib_tip

    c1 = add_tapers(c0, cross_section2=xs_rib_tip, linear=True)
    c1.show()



from typing import Optional, Tuple

import numpy as np

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.array import array
from gdsfactory.components.rectangle import rectangle
from gdsfactory.types import Float2, Layer


@cell
def rectangle_with_slits(
    size: Tuple[float, float] = (100.0, 200.0),
    layer: Layer = (1, 0),
    layer_slit: Optional[Layer] = (2, 0),
    centered: bool = False,
    port_type: Optional[str] = None,
    slit_size: Tuple[float, float] = (1.0, 1.0),
    slit_spacing: Float2 = (20, 20),
    slit_enclosure: float = 10,
) -> Component:
    """Returns a rectangle with slits. Metal slits reduce stress.

    Args:
        size: (tuple) Width and height of rectangle.
        layer: Specific layer to put polygon geometry on.
        layer_slit: does a boolan NOT when None.
        centered: True sets center to (0, 0), False sets south-west to (0, 0)
        port_type: for the rectangle
        slit_size: x, y slit size
        slit_spacing: pitch_x, pitch_y for slits
        slit_enclosure: from slit to rectangle edge


    .. code::

        slit_enclosure
        _____________________________________
        |<--->                              |
        |                                   |
        |      ______________________       |
        |     |                      |      |
        |     |                      | slit_size[1]
        |     |______________________|      |
        |  |                                |
        |  | slit_spacing                   |
        |  |                                |  size[1]
        |  |   ______________________       |
        |  |  |                      |      |
        |  |  |                      |      |
        |  |  |______________________|      |
        |     <--------------------->       |
        |            slit_size[0]           |
        |___________________________________|
                        size[0]


    """
    c = Component()
    r = rectangle(size=size, layer=layer, port_type=port_type, centered=centered)
    c.add_ports(r.ports)
    slit = rectangle(size=slit_size, port_type=None, layer=layer_slit or layer)

    columns = np.floor((size[0] - 2 * slit_enclosure) / slit_spacing[0])
    rows = np.floor((size[1] - 2 * slit_enclosure) / slit_spacing[1])
    slits = array(slit, columns=columns, rows=rows, spacing=slit_spacing).ref()
    slits.xmin = slit_enclosure
    slits.ymin = slit_enclosure

    if layer_slit:
        c << r
        c.add(slits)
    else:
        r_with_slits = c << gf.geometry.boolean(r, slits, operation="not", layer=layer)
        c.absorb(r_with_slits)
    return c


if __name__ == "__main__":
    # c = rectangle_with_slits(layer_slit=None)
    c = rectangle_with_slits(layer_slit=(2, 0), slit_size=(10, 10))
    c.show()



import pytest

import gdsfactory as gf
from gdsfactory.component import MutabilityError


def test_cache_fail(padding: float = 3.0) -> gf.Component:
    """
    Adding padding to a cached component should raise MutabilityError

    Args:
        default: default padding on all sides
    """
    c = gf.c.straight()

    with pytest.raises(MutabilityError):
        c.add_padding(default=padding)  # add padding to original cell

    return c


if __name__ == "__main__":
    c = test_cache_fail()
    c.show()



import gdsfactory as gf


def test_snap_to_grid() -> None:
    assert gf.snap.snap_to_grid(1.1e-3) == 0.001


def test_snap_to_2nm_grid() -> None:
    assert gf.snap.snap_to_2nm_grid(1.1e-3) == 0.002
    assert gf.snap.snap_to_2nm_grid(3.1e-3) == 0.004


def test_is_on_1nm_grid() -> None:
    assert not gf.snap.is_on_grid(0.1e-3)
    assert gf.snap.is_on_grid(1e-3)


def test_is_on_2nm_grid() -> None:
    assert not gf.snap.is_on_grid(1.1e-3, 2)
    assert not gf.snap.is_on_grid(1e-3, 2)
    assert gf.snap.is_on_grid(2e-3, 2)


if __name__ == "__main__":
    test_is_on_2nm_grid()
    # print(snap_to_grid(1.1e-3))
    # print(snap_to_2nm_grid(1.1e-3))
    # print(snap_to_2nm_grid(3.1e-3))

    # print(on_1nm_grid(1.1e-3))
    # print(on_1nm_grid(1e-3))

    # print(on_2nm_grid(1.1e-3))
    # print(on_2nm_grid(1e-3))
    # print(on_2nm_grid(2e-3))



import gdsfactory as gf


def test_route():
    c = gf.Component("sample_connect")
    mmi1 = c << gf.components.mmi1x2()
    mmi2 = c << gf.components.mmi1x2()
    mmi2.move((100, 50))
    route = gf.routing.get_route(
        mmi1.ports["o3"], mmi2.ports["o1"], cross_section=gf.cross_section.metal1
    )
    c.add(route.references)
    c.show()


if __name__ == "__main__":
    c = gf.Component("sample_connect")
    mmi1 = c << gf.components.mmi1x2()
    mmi2 = c << gf.components.mmi1x2()
    mmi2.move((100, 50))
    route = gf.routing.get_route(
        mmi1.ports["o3"], mmi2.ports["o1"], cross_section=gf.cross_section.metal1
    )
    c.add(route.references)
    c.show()



import meep as mp

from gdsfactory.config import logger
from gdsfactory.simulation import plot
from gdsfactory.simulation.get_sparameters_path import get_sparameters_data_meep
from gdsfactory.simulation.gmeep import port_symmetries
from gdsfactory.simulation.gmeep.get_simulation import get_simulation
from gdsfactory.simulation.gmeep.write_sparameters_meep import write_sparameters_meep
from gdsfactory.simulation.gmeep.write_sparameters_meep_mpi import (
    write_sparameters_meep_mpi,
)
from gdsfactory.simulation.gmeep.write_sparameters_meep_mpi_pool import (
    write_sparameters_meep_mpi_pool,
)

logger.info(f"Found Meep {mp.__version__!r} installed at {mp.__path__!r}")

__all__ = [
    "get_simulation",
    "get_sparameters_data_meep",
    "write_sparameters_meep",
    "write_sparameters_meep_mpi",
    "write_sparameters_meep_mpi_pool",
    "plot",
    "port_symmetries",
]
__version__ = "0.0.3"
